"""This module defines the Normalizer class, which handles the normalization of
images and tags based on a given configuration.

The class includes methods to precompile normalization functions,
normalize images and tags, and define specific normalization strategies
such as Z-score and uniform normalization. The normalization process
involves replacing NaN values, creating masks, and scaling values to a
specified range. The module also provides functions to recover the
original values from the normalized data. The Normalizer class ensures
that both images and tags are consistently normalized according to the
specified configuration, facilitating further processing and analysis.
"""
from __future__ import annotations

from typing import Callable

import numpy as np
import torch


class Normalizer:
    """Class to handle the normalization of images and tags.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    """

    def __init__(self, conf: dict):
        self.conf = conf

    def _normalizing_functions(
        self,
        all_images: list[torch.tensor],
        all_tags: list[np.ndarray],
        all_ensemble_ids: list[int],
    ):
        """Precompiles the normalizing functions.

        Parameters
        ----------
        all_images : list
            List of all images
        all_tags : list
            List of all tags
        conf : dict
            Dictionary containing the configuration
        """
        # Define the normalization functions for the images
        if not hasattr(self, 'normalize_img'):
            # First we replace all NaN values by 0
            all_images = [torch.nan_to_num(image) for image in all_images]
            self._mask_img = 1 - np.isnan(all_images[0])
            # Then we create a mask corresponding to all the 0s of image 0
            self._mask_img = self._mask_img * (all_images[0] != 0)
            # as a sanity check, we check if the mask would the same using the last image
            assert np.array_equal(
                self._mask_img, self._mask_img *
                (all_images[-1] != 0)), 'Error: mask is not consistent'
            self._define_img_normalize_functions(all_images)

        # Define the normalization functions for the tags
        if not hasattr(self, 'normalize_tag'):
            self._define_tag_normalize_functions(all_tags)

    def normalize_imgs_and_tags(
        self,
        all_images: list[torch.tensor],
        all_tags: list[np.ndarray],
        all_ensemble_ids: list[int],
    ) -> list[tuple[torch.tensor, np.ndarray, int]]:
        """Normalize both the input images and the tags to be between 0 and 1.

        Parameters
        ----------
        all_images : list
            List of all images
        all_tags : list
            List of all tags
        conf : dict
            Dictionary containing the configuration

        Returns
        -------
        dataset : list
            List of tuples (image, normalized_tag)
        """
        self._normalizing_functions(all_images, all_tags, all_ensemble_ids)

        # Normalize the images
        normalized_images = [self.normalize_img(image) for image in all_images]

        # And normalize the tags
        normalized_tags = np.array([[
            self.normalize_tag[j](tag, tag_id) for j, tag in enumerate(tags)
        ] for tag_id, tags in enumerate(all_tags)],
                                   dtype=np.float32)

        # I can now create the dataset
        dataset = [(image, tag, ensemble_id) for image, tag, ensemble_id in
                   zip(normalized_images, normalized_tags, all_ensemble_ids)]

        return dataset

    def _define_img_normalize_functions(self, all_images):
        """Define the normalization functions for the images."""
        if self.conf['preproc']['img_normalization']:
            # Find the maximum between the maximum and minimum values of the images
            max_img = max([torch.max(image) for image in all_images])
            print('*** Image max:', max_img)
            min_img = min([torch.min(image) for image in all_images])
            print('*** Image min:', min_img)
            range_img = max_img - min_img

            self.normalize_img, self.recover_img = self._img_normalize_functions(
                min_img, range_img)
        else:
            # The images do not need to be normalized
            print('-> [!] No img normalization.')
            self.normalize_img = lambda x: x
            self.recover_img = lambda x, y: x

    def _define_tag_normalize_functions(self, all_tags):
        """Define the normalization functions for the tags."""
        self.min_tags = np.min(all_tags, axis=0)
        print('*** Tag min:', self.min_tags)
        self.max_tags = np.max(all_tags, axis=0)
        print('*** Tag max:', self.max_tags)

        # get the std deviation if using Z-score
        if self.conf['preproc']['normalization'] == 'zscore':
            self._define_zscore_normalize_functions(all_tags)
        elif self.conf['preproc']['normalization'] == 'unif':
            self._define_unif_normalize_functions(all_tags)

        self.normalize_tag, self.recover_tag = self._tag_normalize_functions()

    def _define_zscore_normalize_functions(self, all_tags):
        """Define the normalization functions for the tags using Z-score."""
        self.mean_tags = np.mean(all_tags, axis=0)
        print('*** Tag mean:', self.mean_tags)
        self.std_tags = np.std(all_tags, axis=0)
        print('*** Tag std:', self.std_tags)

    def _define_unif_normalize_functions(self, all_tags):
        """Define the normalization functions for the tags using uniform
        normalization."""
        raise NotImplementedError(
            '*** uniform normalization is not fully implemented yet. The sorting messes up the ensemble IDs.'
        )
        array_tags = np.array(all_tags)

        # Find the indices that sort the tags
        _sorting_indices = np.argsort(array_tags, axis=0)
        # And the corresponding indices that unsort them
        self._unsorting_indices = np.argsort(np.argsort(array_tags, axis=0),
                                             axis=0)
        self._sorted_tags = np.stack([
            array_tags[_sorting_indices[:, i], i]
            for i in range(array_tags.shape[1])
        ]).T
        self.Ndata = array_tags.shape[0]

    def _img_normalize_functions(
            self, min_img: np.ndarray,
            range_img: np.ndarray) -> tuple[Callable, Callable]:
        """Create the normalization and recovery functions for the images. The
        images are normalized between 0 and 1 using global values.

        Parameters
        ----------
        min_img : np.ndarray
            Minimum value for all the images
        range_img : np.ndarray
            Maximum value for all the images

        Returns
        -------
        normalize_fn : Callable
            Function to normalize an image
        recover_fn : Callable
            Function to recover an image
        """

        def normalize_fn(x, min_img=min_img, range_img=range_img):
            return self._mask_img * (x - min_img) / range_img

        def recover_fn(y, y_id, min_img=min_img, range_img=range_img):
            return self._mask_img * (y * range_img + min_img)

        return normalize_fn, recover_fn

    def _tag_normalize_functions(
            self) -> tuple[list[Callable], list[Callable]]:
        """Create the normalization and recovery functions for the tags.
        Several alternatives are available.

        Returns
        -------
        normalize_functions : list
            List of functions to normalize each tag
        recover_functions : list
            List of functions to recover each tag
        """
        normalization = self.conf['preproc']['normalization']
        nscreens = self.conf['speckle']['nscreens']

        if normalization == 'unif':
            return self._unif_normalize_functions(nscreens)
        elif normalization == 'zscore':
            return self._zscore_normalize_functions(nscreens)
        elif normalization == 'log':
            return self._log_normalize_functions(nscreens)
        elif normalization == 'lin':
            return self._lin_normalize_functions(nscreens)
        else:
            raise ValueError(
                f'*** Error in normalization: normalization {normalization} unknown.'
            )

    def _unif_normalize_functions(self, nscreens):

        def create_normalize_functions(nscreens):

            def normalize_fn(x, x_id, i):
                return self._unsorting_indices[x_id, i] / self.Ndata

            return [normalize_fn for _ in range(nscreens)]

        def create_recover_functions(nscreens):

            def recover_fn(y, y_id, i):
                return self._sorted_tags[round(y * self.Ndata), i]

            return [recover_fn for _ in range(nscreens)]

        normalize_functions = create_normalize_functions(nscreens)
        recover_functions = create_recover_functions(nscreens)
        return normalize_functions, recover_functions

    def _zscore_normalize_functions(self, nscreens):
        # TODO if you still decide to support this,
        # return def functions and not lambdas
        normalize_functions = [
            (lambda x, x_id, mean=self.mean_tags[i], std=self.std_tags[i]:
             (x - mean) / std) for i in range(nscreens)
        ]
        recover_functions = [(lambda y, y_id, mean=self.mean_tags[
            i], std=self.std_tags[i]: y * std + mean) for i in range(nscreens)]
        return normalize_functions, recover_functions

    def _log_normalize_functions(self, nscreens):
        # TODO if you still decide to support this,
        # return def functions and not lambdas
        normalize_functions = [(lambda x, x_id, min_t=self.min_tags[
            i], max_t=self.max_tags[i]: np.log(
                (x - min_t + 1) / (max_t - min_t + 1)) / np.log(
                    (2 - min_t) / (max_t - min_t + 1)))
                               for i in range(nscreens)]
        recover_functions = [(lambda y, y_id, min_t=self.min_tags[
            i], max_t=self.max_tags[i]: np.exp(y) * np.log(
                (2 - min_t) / (max_t - min_t + 1)) *
                              (max_t - min_t + 1) + min_t - 1)
                             for i in range(nscreens)]
        return normalize_functions, recover_functions

    def _lin_normalize_functions(self, nscreens):
        # TODO if you still decide to support this,
        # return def functions and not lambdas
        normalize_functions = [
            (lambda x, x_id, min_t=self.min_tags[i], max_t=self.max_tags[i]:
             (x - min_t) / (max_t - min_t)) for i in range(nscreens)
        ]
        recover_functions = [(
            lambda y, y_id, min_t=self.min_tags[i], max_t=self.max_tags[i]: y *
            (max_t - min_t) + min_t) for i in range(nscreens)]
        return normalize_functions, recover_functions



from __future__ import annotations

import random
import time

import torch
from torch import nn, optim

from speckcn2.io import save
from speckcn2.loss import ComposableLoss
from speckcn2.mlmodels import EarlyStopper, EnsembleModel
from speckcn2.plots import score_plot
from speckcn2.preprocess import Normalizer
from speckcn2.utils import ensure_directory


def train(model: nn.Module, last_model_state: int, conf: dict, train_set: list,
          test_set: list, device: torch.device, optimizer: optim.Optimizer,
          criterion: ComposableLoss,
          criterion_val: ComposableLoss) -> tuple[nn.Module, float]:
    """Trains the model for the given number of epochs.

    Parameters
    ----------
    model : torch.nn.Module
        The model to train
    last_model_state : int
        The number of the last model state
    conf : dict
        Dictionary containing the configuration
    train_set : list
        The training set
    test_set : list
        The testing set
    device : torch.device
        The device to use
    optimizer : torch.optim
        The optimizer to use
    criterion : ComposableLoss
        The loss function to use
    criterion_val : ComposableLoss
        The loss function to use for validation

    Returns
    -------
    model : torch.nn.Module
        The trained model
    average_loss : float
        The average loss of the last epoch
    """

    final_epoch = conf['hyppar']['maxepochs']
    save_every = conf['model']['save_every']
    datadirectory = conf['speckle']['datadirectory']
    batch_size = conf['hyppar']['batch_size']

    if getattr(model, 'early_stop', False):
        print(
            'Warning: Training reached early stop in a previous training instance'
        )
        return model, 0

    print(f'Training the model from epoch {last_model_state} to {final_epoch}')

    # Setup the EnsembleModel wrapper
    ensemble = EnsembleModel(conf, device)

    # Early stopper
    early_stopping = conf['hyppar'].get('early_stopping', -1)
    if early_stopping > 0:
        print('Using early stopping (patience = {})'.format(early_stopping))
        min_delta = conf['hyppar'].get('early_stop_delta', 0.1)
        early_stopper = EarlyStopper(patience=early_stopping,
                                     min_delta=min_delta)

    average_loss = 0.0
    model.train()
    for epoch in range(last_model_state, final_epoch):
        total_loss = 0.0
        model.train()
        t_in = time.time()
        for i in range(0, len(train_set), batch_size):
            batch = train_set[i:i + batch_size]

            # Zero out the optimizer
            optimizer.zero_grad()

            # Forward pass
            outputs, targets, _ = ensemble(model, batch)
            loss, _ = criterion(outputs, targets)

            # Backward pass
            loss.backward()
            optimizer.step()

            # Accumulate the loss
            total_loss += loss.item()

        # Shuffle the training set
        random.shuffle(train_set)

        # Calculate average loss for the epoch
        average_loss = total_loss / len(train_set)

        # Log the important information
        t_fin = time.time() - t_in
        model.loss.append(average_loss)
        model.time.append(t_fin)
        model.epoch.append(epoch + 1)

        # And also the validation loss
        val_loss = 0.0
        model.eval()
        with torch.no_grad():
            for i in range(0, len(test_set), batch_size):
                batch = test_set[i:i + batch_size]
                # Forward pass
                outputs, targets, _ = ensemble(model, batch)
                loss, _ = criterion_val(outputs, targets)
                # sum loss
                val_loss += loss.item()
        val_loss = val_loss / len(test_set)
        model.val_loss.append(val_loss)

        # Print the average loss for every epoch
        message = (f'Epoch {epoch+1}/{final_epoch} '
                   f'(in {t_fin:.3g}s),\tTrain-Loss: {average_loss:.5f},\t'
                   f'Test-Loss: {val_loss:.5f}')
        print(message, flush=True)

        if early_stopper.early_stop(val_loss):
            print('Early stopping triggered')
            save(model, datadirectory, early_stop=True)

        if (epoch + 1) % save_every == 0 or epoch == final_epoch - 1:
            save(model, datadirectory)

    return model, average_loss


def score(
        model: nn.Module,
        test_set: list,
        device: torch.device,
        criterion: ComposableLoss,
        normalizer: Normalizer,
        nimg_plot: int = 100
) -> tuple[list, list, list, list, list, list, list]:
    """Tests the model.

    Parameters
    ----------
    model : torch.nn.Module
        The model to test
    test_set : list
        The testing set
    device : torch.device
        The device to use
    criterion : ComposableLoss
        The composable loss function, where I can access useful parameters
    normalizer : Normalizer
        The normalizer used to recover the tags
    nimg_plot : int
        Number of images to plot

    Returns
    -------
    test_tags : list
        List of all the predicted tags of the test set
    test_losses : list
        List of all the losses of the test set
    test_measures : list
        List of all the measures of the test set
    test_cn2_pred : list
        List of all the predicted Cn2 profiles of the test set
    test_cn2_true : list
        List of all the true Cn2 profiles of the test set
    test_recovered_tag_pred : list
        List of all the recovered tags from the model prediction
    test_recovered_tag_true : list
        List of all the recovered tags
    """
    counter = 0
    conf = normalizer.conf
    data_dir = conf['speckle']['datadirectory']
    batch_size = conf['hyppar']['batch_size']
    # Setup the EnsembleModel wrapper
    ensemble = EnsembleModel(conf, device)

    # For scoring the model, it is possible to compose the loss
    # in the same way as you did during training adn validation.
    # However, at the moment we are forcing the computation
    # of all the quantities of interest.
    # This can be changed in the future to save a bit of time.
    criterion.loss_weights = {
        'MSE': 1,
        'MAE': 1,
        'JMSE': 0,
        'JMAE': 0,
        'Cn2MSE': 0,
        'Cn2MAE': 0,
        'Pearson': 0,
        'Fried': 1,
        'Isoplanatic': 1,
        'Rytov': 0,
        'Scintillation_w': 1,
        'Scintillation_ms': 0,
    }
    criterion._select_loss_needed()

    with torch.no_grad():
        # Put model in evaluation mode
        model.eval()

        test_tags = []
        test_losses = []
        test_measures = []
        test_cn2_pred = []
        test_cn2_true = []
        test_recovered_tag_pred = []
        test_recovered_tag_true = []
        # Initialize the loss max and min. They are used to plot the images with the
        # highest and lowest loss. We skip examples with a common average value of the loss.
        loss_max = 0
        loss_min = 1e6
        # create the directory where the images will be stored
        ensure_directory(f'{data_dir}/{model.name}_score')

        for idx in range(0, len(test_set), batch_size):
            batch = test_set[idx:idx + batch_size]

            # Forward pass
            outputs, targets, inputs = ensemble(model, batch)

            # Loop each input separately
            for i in range(len(outputs)):
                loss, losses = criterion(outputs[i], targets[i])

                # Get the Cn2 profile and the recovered tags
                Cn2_pred = criterion.reconstruct_cn2(outputs[i])
                Cn2_true = criterion.reconstruct_cn2(targets[i])
                recovered_tag_pred = criterion.get_J(outputs[i])
                recovered_tag_true = criterion.get_J(targets[i])
                # and get all the measures
                all_measures = criterion._get_all_measures(
                    outputs[i], targets[i], Cn2_pred, Cn2_true)
                this_loss = loss.item()

                if counter < nimg_plot and (this_loss > loss_max
                                            or this_loss < loss_min):
                    loss_max = max(this_loss, loss_max)
                    loss_min = min(this_loss, loss_min)
                    print(f'Plotting item {counter} loss: {this_loss:.4f}')
                    score_plot(conf, inputs, targets, loss, losses, i, counter,
                               all_measures, Cn2_pred, Cn2_true,
                               recovered_tag_pred, recovered_tag_true)
                    counter += 1

                # and get all the tags for statistic analysis
                for tag in outputs:
                    test_tags.append(tag)

                # and get the other information
                test_losses.append(losses)
                test_measures.append(all_measures)
                test_cn2_pred.append(Cn2_pred)
                test_cn2_true.append(Cn2_true)
                test_recovered_tag_pred.append(recovered_tag_pred)
                test_recovered_tag_true.append(recovered_tag_true)

    return (test_tags, test_losses, test_measures, test_cn2_pred,
            test_cn2_true, test_recovered_tag_pred, test_recovered_tag_true)



from __future__ import annotations

import itertools
from typing import Callable, Dict, Optional

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import torch
from scipy import stats
from torch import Tensor, nn
from torch import device as Device

from speckcn2.loss import ComposableLoss
from speckcn2.mlmodels import EnsembleModel
from speckcn2.utils import ensure_directory


def tags_distribution(conf: dict,
                      train_set: list,
                      test_tags: Tensor,
                      device: Device,
                      nbins: int = 20,
                      rescale: bool = False,
                      recover_tag: Optional[list[Callable]] = None) -> None:
    """Function to plot the following:
    - distribution of the tags for unscaled results
    - distribution of the tags for rescaled results
    - distribution of the sum of the tags.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    train_set : list
        The training set
    test_tags : torch.Tensor
        The predicted tags for the test dataset
    device : torch.device
        The device to use
    data_directory : str
        The directory where the data is stored
    nbins : int, optional
        Number of bins to use for the histograms
    rescale : bool, optional
        Whether to rescale the tags using recover_tag() or leave them between 0 and 1
    recover_tag : list, optional
        List of functions to recover each tag
    """

    data_directory = conf['speckle']['datadirectory']
    model_name = conf['model']['name']
    ensemble = conf['preproc'].get('ensemble', 1)

    dirname = f'{data_directory}/{model_name}_score'
    ensure_directory(dirname)

    # Get the tags from the training set
    if ensemble > 1:
        train_set = list(itertools.chain(*train_set))
    _, tags, _ = zip(*train_set)
    tags = np.stack(tags)
    train_tags = np.array([n for n in tags])

    # Get the tags from the test set
    predic_tags = np.array([n.cpu().numpy() for n in test_tags])

    # Keep track of J=sum(tags) for each sample
    J_pred = np.zeros(predic_tags.shape[0])
    J_true = np.zeros(train_tags.shape[0])

    # Plot the distribution of each tag element
    fig, axs = plt.subplots(2, 4, figsize=(20, 10))
    for i in range(train_tags.shape[1]):
        if rescale and recover_tag is not None:
            recovered_tag_model = np.asarray(
                [recover_tag[i](predic_tags[:, i], i)]).squeeze(0)
            recovered_tag_true = np.asarray(
                [recover_tag[i](train_tags[:, i], i)]).squeeze(0)
            J_pred += 10**recovered_tag_model
            J_true += 10**recovered_tag_true
            axs[i // 4, i % 4].hist(recovered_tag_model,
                                    bins=nbins,
                                    color='tab:red',
                                    density=True,
                                    alpha=0.5,
                                    label='Model prediction')
            axs[i // 4, i % 4].hist(recovered_tag_true,
                                    bins=nbins,
                                    color='tab:blue',
                                    density=True,
                                    alpha=0.5,
                                    label='Training data')
        else:
            axs[i // 4, i % 4].hist(predic_tags[:, i],
                                    bins=nbins,
                                    color='tab:red',
                                    density=True,
                                    alpha=0.5,
                                    label='Model prediction')
            axs[i // 4, i % 4].hist(train_tags[:, i],
                                    bins=nbins,
                                    color='tab:blue',
                                    density=True,
                                    alpha=0.5,
                                    label='Training data')
        axs[i // 4, i % 4].set_title(f'Screen {i}')
    axs[0, 1].legend()
    plt.tight_layout()
    figname = f'{dirname}/{model_name}_tags'
    if not rescale:
        figname += '_unscaled'
    plt.savefig(f'{figname}.png')
    plt.close()

    if rescale and recover_tag is not None:
        # Also plot the distribution of the sum of the tags
        fig, axs = plt.subplots(1, 1, figsize=(6, 6))
        axs.hist(np.log10(J_pred),
                 bins=nbins,
                 color='tab:red',
                 density=True,
                 alpha=0.5,
                 label='Model prediction')
        axs.hist(np.log10(J_true),
                 bins=nbins,
                 color='tab:blue',
                 density=True,
                 alpha=0.5,
                 label='Training data')
        axs.set_title('Sum of J')
        axs.legend()
        plt.tight_layout()
        plt.savefig(f'{dirname}/{model_name}_sumJ.png')

        plt.close()


def average_speckle_output(conf: dict,
                           test_set: list,
                           device: Device,
                           model: nn.Torch,
                           criterion: ComposableLoss,
                           trimming: float = 0.1,
                           n_ensembles_to_plot: int = 100) -> None:
    """Test to see if averaging the prediction of multiple speckle patterns
    improves the results. This function is then going to plot the relative
    error over the screen tags and the Fried parameter to make this evaluation.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    test_set : list
        The test set
    device : torch.device
        The device to use
    model : nn.Torch
        The trained model
    criterion : ComposableLoss
        The loss function
    n_ensembles_to_plot : int
        The number of ensembles to plot
    """

    data_directory = conf['speckle']['datadirectory']
    model_name = conf['model']['name']
    n_screens = conf['speckle']['nscreens']

    dirname = f'{data_directory}/{model_name}_score/effect_averaging'
    ensure_directory(dirname)

    # group the sets that have the same n[1]
    grouped_test_set: Dict = {}
    for n in test_set:
        key = tuple(n[1])
        if key not in grouped_test_set:
            grouped_test_set[key] = []
        grouped_test_set[key].append(n)
    print('\nChecking if averaging speckle predictions improves results')
    print(f'Number of samples: {len(test_set)}')
    print(f'Number of speckle groups: {len(grouped_test_set)}')

    # In the end, we will plot groups that have uncommon values of loss
    loss_min = 1e10
    loss_max = 0
    ensemble_count = 0

    ensemble = EnsembleModel(conf, device)
    with torch.no_grad():
        model.eval()

        for key, value in grouped_test_set.items():
            _outputs = []
            _all_tags_pred = []
            cmap = plt.get_cmap('coolwarm')
            norm = plt.Normalize(1, len(value))

            for count, speckle in enumerate(value, 1):
                color = cmap(norm(count))
                output, target, _ = ensemble(model, [speckle])

                if count == 1:
                    fig, ax = plt.subplots(1, 4, figsize=(16, 4))

                    # (1) Plot J vs nscreens
                    recovered_tag_true = criterion.get_J(target)
                    ax[0].plot(recovered_tag_true.squeeze(0).detach().cpu(),
                               '*',
                               label='True',
                               color='tab:green',
                               markersize=10,
                               markeredgecolor='black',
                               zorder=100)
                    ax[1].plot(recovered_tag_true.squeeze(0).detach().cpu(),
                               '*',
                               label='True',
                               color='tab:green',
                               markersize=10,
                               markeredgecolor='black',
                               zorder=100)
                    blue_patch = mpatches.Patch(color=color,
                                                label='One speckle')

                # Use the trimmed mean to get the average output
                # (trim_mean works only on cpu so you have to move back and forth)
                _outputs.append(output.detach().cpu().numpy())
                avg_output = torch.tensor(stats.trim_mean(_outputs,
                                                          trimming)).to(device)

                loss, losses = criterion(avg_output, target)

                # Get the Cn2 profile and the recovered tags
                Cn2_pred = criterion.reconstruct_cn2(avg_output)
                Cn2_true = criterion.reconstruct_cn2(target)
                recovered_tag_pred = criterion.get_J(avg_output)
                _all_tags_pred.append(recovered_tag_pred.detach().cpu())
                ax[1].plot(recovered_tag_pred.squeeze(0).detach().cpu(),
                           'o',
                           color=color)
                # and get all the measures
                all_measures = criterion._get_all_measures(
                    avg_output, target, Cn2_pred, Cn2_true)

                Fried_err = torch.abs(
                    all_measures['Fried_true'] -
                    all_measures['Fried_pred']) / all_measures['Fried_true']
                ax[2].scatter(count, Fried_err.detach().cpu(), color=color)
                if count == 1:
                    ax[2].plot(
                        [], [],
                        ' ',
                        label='(True) Fried = {:.3f}'.format(
                            all_measures['Fried_true'].detach().cpu().numpy()))

            # Now at the end of the loop, we decide if this set needs to plotted or not
            # by checking that the loss
            if loss > loss_max or loss < loss_min:
                avg_tags_trim = stats.trim_mean(_all_tags_pred,
                                                trimming).squeeze()
                percentiles_50 = np.percentile(_all_tags_pred, [25, 75],
                                               axis=0).squeeze()
                percentiles_68 = np.percentile(_all_tags_pred, [16, 84],
                                               axis=0).squeeze()
                percentiles_95 = np.percentile(_all_tags_pred, [2.5, 97.5],
                                               axis=0).squeeze()

                x_vals = np.arange(n_screens)
                alp = 1
                ax[0].plot(avg_tags_trim,
                           label='Mean',
                           color='tab:red',
                           zorder=50)
                ax[0].fill_between(x_vals,
                                   percentiles_50[0],
                                   percentiles_50[1],
                                   color='gold',
                                   alpha=alp,
                                   label='50% CI',
                                   zorder=5)
                ax[0].fill_between(x_vals,
                                   percentiles_68[0],
                                   percentiles_50[0],
                                   color='cadetblue',
                                   alpha=alp,
                                   label='68% CI',
                                   zorder=4)
                ax[0].fill_between(x_vals,
                                   percentiles_50[1],
                                   percentiles_68[1],
                                   color='cadetblue',
                                   alpha=alp,
                                   zorder=4)
                ax[0].fill_between(x_vals,
                                   percentiles_95[0],
                                   percentiles_68[0],
                                   color='blue',
                                   label='95% CI',
                                   alpha=alp,
                                   zorder=3)
                ax[0].fill_between(x_vals,
                                   percentiles_68[1],
                                   percentiles_95[1],
                                   color='blue',
                                   alpha=alp,
                                   zorder=3)

                ax[3].axis('off')  # Hide axis
                recap_info = f'LOSS TERMS:\nTotal Loss: {loss.item():.4g}\n'
                # the individual losses
                for key, value in losses.items():
                    recap_info += f'{key}: {value.item():.4g}\n'
                recap_info += '-------------------\nPARAMETERS:\n'
                # then the single parameters
                for key, value in all_measures.items():
                    recap_info += f'{key}: {value:.4g}\n'
                ax[3].text(0.5,
                           0.5,
                           recap_info,
                           horizontalalignment='center',
                           verticalalignment='center',
                           fontsize=10,
                           color='black')

                ax[0].set_yscale('log')
                ax[0].set_ylabel('J')
                ax[0].set_xlabel('# screen')
                ax[0].legend()
                ax[1].set_yscale('log')
                ax[1].set_ylabel('J')
                ax[1].set_xlabel('# screen')
                red_patch = mpatches.Patch(color=color, label='All speckles')
                handles, labels = ax[1].get_legend_handles_labels()
                handles.extend([blue_patch, red_patch])
                ax[1].legend(handles=handles)
                ax[2].set_xlabel('# averaged speckles')
                ax[2].set_ylabel('Fried relative error')
                ax[2].legend(frameon=False)
                ax[2].set_yscale('log')
                fig.tight_layout()
                plt.subplots_adjust(top=0.92)
                plt.suptitle('Effect of averaging speckle predictions')
                plt.savefig(
                    f'{dirname}/average_ensemble_loss{loss.item():.4g}.png')
                loss_max = max(loss, loss_max)
                loss_min = min(loss, loss_min)
                ensemble_count += 1

            plt.close()

            if ensemble_count > n_ensembles_to_plot:
                break


def average_speckle_input(conf: dict,
                          test_set: list,
                          device: Device,
                          model: nn.Torch,
                          criterion: ComposableLoss,
                          n_ensembles_to_plot: int = 100) -> None:
    """Test to see if averaging the speckle patterns (before the prediction)
    improves the results. This function is then going to plot the relative
    error over the screen tags and the Fried parameter to make this evaluation.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    test_set : list
        The test set
    device : torch.device
        The device to use
    model : nn.Torch
        The trained model
    criterion : ComposableLoss
        The loss function
    n_ensembles_to_plot : int
        The number of ensembles to plot
    """

    data_directory = conf['speckle']['datadirectory']
    model_name = conf['model']['name']

    dirname = f'{data_directory}/{model_name}_score/effect_averaging'
    ensure_directory(dirname)

    # group the sets that have the same n[1]
    grouped_test_set: Dict = {}
    for n in test_set:
        key = tuple(n[1])
        if key not in grouped_test_set:
            grouped_test_set[key] = []
        grouped_test_set[key].append(n)
    print('\nChecking if averaging speckle patterns improves results')
    print(f'Number of samples: {len(test_set)}')
    print(f'Number of speckle groups: {len(grouped_test_set)}')

    # For each group compare the model prediction to the exact tag
    ensemble = EnsembleModel(conf, device)
    with torch.no_grad():
        model.eval()

        for ensemble_count, (key,
                             value) in enumerate(grouped_test_set.items()):
            avg_speckle = None
            cmap = plt.get_cmap('coolwarm')
            norm = plt.Normalize(1, len(value))

            if ensemble_count > n_ensembles_to_plot:
                continue

            for count, speckle in enumerate(value, 1):
                color = cmap(norm(count))

                if avg_speckle is None:
                    avg_speckle = speckle
                else:
                    avg_speckle = (torch.add(avg_speckle[0],
                                             speckle[0]), *avg_speckle[1:])

                # Average only the speckle pattern (first element)
                avg_speckle_divided = (torch.div(avg_speckle[0],
                                                 count), *avg_speckle[1:])

                output, target, _ = ensemble(model, [avg_speckle_divided])
                loss, losses = criterion(output, target)

                if count == 1:
                    fig, ax = plt.subplots(1, 3, figsize=(12, 4))

                    recovered_tag_true = criterion.get_J(target)
                    ax[0].plot(recovered_tag_true.squeeze(0).detach().cpu(),
                               '*',
                               label='True',
                               color='tab:green',
                               zorder=100)
                    blue_patch = mpatches.Patch(color=color,
                                                label='One speckle')

                ax[1].plot((torch.abs(output - target) /
                            (target + 1e-7)).flatten().detach().cpu(),
                           color=color)

                # Get the Cn2 profile and the recovered tags
                Cn2_pred = criterion.reconstruct_cn2(output)
                Cn2_true = criterion.reconstruct_cn2(target)
                recovered_tag_pred = criterion.get_J(output / count)
                ax[0].plot(recovered_tag_pred.squeeze(0).detach().cpu(),
                           '.',
                           color=color)
                # and get all the measures
                all_measures = criterion._get_all_measures(
                    output, target, Cn2_pred, Cn2_true)

                Fried_err = torch.abs(
                    all_measures['Fried_true'] -
                    all_measures['Fried_pred']) / all_measures['Fried_true']
                ax[2].scatter(count, Fried_err.detach().cpu(), color=color)
                if count == 1:
                    ax[2].plot(
                        [], [],
                        ' ',
                        label='(True) Fried = {:.3f}'.format(
                            all_measures['Fried_true'].detach().cpu().numpy()))

            ax[0].set_yscale('log')
            ax[0].set_ylabel('J')
            ax[0].set_xlabel('# screen')
            red_patch = mpatches.Patch(color=color, label='All speckles')
            handles, labels = ax[0].get_legend_handles_labels()
            handles.extend([blue_patch, red_patch])
            ax[0].legend(handles=handles)
            ax[1].set_xlabel('# screen')
            ax[1].set_ylabel('J relative error ')
            ax[2].set_xlabel('# averaged speckles')
            ax[2].set_ylabel('Fried relative error')
            ax[1].set_yscale('log')
            ax[2].set_yscale('log')
            ax[2].legend(frameon=False)
            fig.tight_layout()
            plt.subplots_adjust(top=0.92)
            plt.suptitle('Effect of averaging speckle patterns')
            plt.savefig(f'{dirname}/average_speckle_loss{loss.item():.4g}.png')
            plt.close()


def screen_errors(conf: dict,
                  device: Device,
                  J_pred: Tensor,
                  J_true: Tensor,
                  nbins: int = 20,
                  trimming: float = 0.1) -> None:
    """Plot the relative error of Cn2 for each screen.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    device : torch.device
        The device on which the data are stored
    J_pred : torch.Tensor
        The predicted J profile
    J_true : torch.Tensor
        The true J profile
    nbins : int, optional
        Number of bins to use for the histograms
    trimming : float, optional
        The fraction of data to trim from each end of the distribution
    """
    data_directory = conf['speckle']['datadirectory']
    model_name = conf['model']['name']
    n_screens = conf['speckle']['nscreens']

    dirname = f'{data_directory}/{model_name}_score'
    ensure_directory(dirname)

    # Plot the distribution of each tag element
    fig, axs = plt.subplots(2, 4, figsize=(20, 10))
    for si in range(n_screens):

        if device == 'cpu':
            data_pred = np.asarray(J_pred)[:, 0, si]
            data_true = np.asarray(J_true)[:, 0, si]
        else:
            data_pred = np.asarray([d.detach().cpu().numpy()
                                    for d in J_pred])[:, 0, si]
            data_true = np.asarray([d.detach().cpu().numpy()
                                    for d in J_true])[:, 0, si]
        # Define the bins
        bins = np.linspace(min(data_true), max(data_true), nbins + 1)

        # Digitize the data to get bin indices
        bin_indices = np.digitize(data_true, bins)

        # Compute the relative error for each bin
        relative_errors = []
        percentiles_50 = []
        percentiles_68 = []
        percentiles_95 = []

        for bi in range(1, len(bins)):
            model_bin_values = data_pred[bin_indices == bi]
            true_bin_values = data_true[bin_indices == bi]

            if len(true_bin_values) > 0:
                relative_error = np.abs(
                    (model_bin_values - true_bin_values) / true_bin_values)
                relative_errors.append(
                    stats.trim_mean(relative_error, trimming))
                percentiles_50.append(
                    np.percentile(relative_error, [25, 75], axis=0))
                percentiles_68.append(
                    np.percentile(relative_error, [16, 84], axis=0))
                percentiles_95.append(
                    np.percentile(relative_error, [2.5, 97.5], axis=0))
            else:
                relative_errors.append(0)
                percentiles_50.append([0, 0])
                percentiles_68.append([0, 0])
                percentiles_95.append([0, 0])

        percentiles_50 = np.asarray(percentiles_50)
        percentiles_68 = np.asarray(percentiles_68)
        percentiles_95 = np.asarray(percentiles_95)
        # Plot the relative errors
        axs[si // 4, si % 4].plot(bins[:-1],
                                  relative_errors,
                                  label='Mean',
                                  color='tab:red',
                                  zorder=50)
        axs[si // 4, si % 4].set_title(f'Screen {si}')
        # and the percentiles
        axs[si // 4, si % 4].fill_between(bins[:-1],
                                          percentiles_50[:, 0],
                                          percentiles_50[:, 1],
                                          color='gold',
                                          alpha=0.5,
                                          label='50% CI',
                                          zorder=5)
        axs[si // 4, si % 4].fill_between(bins[:-1],
                                          percentiles_68[:, 0],
                                          percentiles_50[:, 0],
                                          color='cadetblue',
                                          alpha=0.5,
                                          label='68% CI',
                                          zorder=4)
        axs[si // 4, si % 4].fill_between(bins[:-1],
                                          percentiles_50[:, 1],
                                          percentiles_68[:, 1],
                                          color='cadetblue',
                                          alpha=0.5,
                                          zorder=4)
        axs[si // 4, si % 4].fill_between(bins[:-1],
                                          percentiles_95[:, 0],
                                          percentiles_68[:, 0],
                                          color='blue',
                                          label='95% CI',
                                          alpha=0.5,
                                          zorder=3)
        axs[si // 4, si % 4].fill_between(bins[:-1],
                                          percentiles_68[:, 1],
                                          percentiles_95[:, 1],
                                          color='blue',
                                          alpha=0.5,
                                          zorder=3)
        axs[si // 4, si % 4].set_yscale('symlog', linthresh=0.1)
    axs[0, 1].legend()
    for ax in axs.flatten():
        ax.axhline(
            y=0,
            linestyle='--',
            color='black',
        )
    vals = axs[0, 0].get_yticks()
    axs[0, 0].set_yticklabels(['{:.0f}'.format(x * 100) + '%' for x in vals])
    vals = axs[-1, -1].get_yticks()
    axs[-1, -1].set_yticklabels(['{:.0f}'.format(x * 100) + '%' for x in vals])
    for i in range(1, n_screens - 1):
        axs.flat[i].sharey(axs.flat[0])
    plt.suptitle('Relative Error of J')
    plt.tight_layout()
    figname = f'{dirname}/{model_name}_Jerrors'
    plt.savefig(f'{figname}.png')
    plt.close()



"""This module provides utility functions for image processing and model
optimization.

It includes functions to plot original and preprocessed images along
with their tags, ensure the existence of specified directories, set up
optimizers based on configuration files, and create circular masks with
an inner "spider" circle removed. These utilities facilitate various
tasks in image analysis and machine learning model training.
"""
from __future__ import annotations

import os

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn


def plot_preprocessed_image(image_orig: torch.tensor,
                            image: torch.tensor,
                            tags: torch.tensor,
                            counter: int,
                            datadirectory: str,
                            mname: str,
                            file_name: str,
                            polar: bool = False) -> None:
    """Plots the original and preprocessed image, and the tags.

    Parameters
    ----------
    image_orig : torch.tensor
        The original image
    image : torch.tensor
        The preprocessed image
    tags : torch.tensor
        The screen tags
    counter : int
        The counter of the image
    datadirectory : str
        The directory containing the data
    mname : str
        The name of the model
    file_name : str
        The name of the original image
    polar : bool, optional
        If the image is in polar coordinates, by default False
    """

    fig, axs = plt.subplots(1, 3, figsize=(15, 5))
    # Plot the original image
    axs[0].imshow(image_orig.squeeze(), cmap='bone')
    axs[0].set_title(f'Training Image {file_name}')
    # Plot the preprocessd image
    axs[1].imshow(image.squeeze(), cmap='bone')
    axs[1].set_title('Processed as')
    if polar:
        axs[1].set_xlabel(r'$r$')
        axs[1].set_ylabel(r'$\theta$')

    # Plot the tags
    axs[2].plot(tags, 'o')
    axs[2].set_yscale('log')
    axs[2].set_title('Screen Tags')
    axs[2].legend()

    fig.subplots_adjust(wspace=0.3)
    plt.savefig(f'{datadirectory}/imgs_to_{mname}/{counter}.png')
    plt.close()


def ensure_directory(data_directory: str) -> None:
    """Ensure that the directory exists.

    Parameters
    ----------
    data_directory : str
        The directory to ensure
    """

    if not os.path.isdir(data_directory):
        os.mkdir(data_directory)


def setup_optimizer(config: dict, model: nn.Module) -> nn.Module:
    """Returns the optimizer specified in the configuration file.

    Parameters
    ----------
    config : dict
        Dictionary containing the configuration
    model : torch.nn.Module
        The model to optimize

    Returns
    -------
    optimizer : torch.nn.Module
        The optimizer with the loaded state
    """

    optimizer_name = config['hyppar']['optimizer']
    if optimizer_name == 'Adam':
        return torch.optim.Adam(model.parameters(), lr=config['hyppar']['lr'])
    elif optimizer_name == 'SGD':
        return torch.optim.SGD(model.parameters(), lr=config['hyppar']['lr'])
    else:
        raise ValueError(f'Unknown optimizer {optimizer_name}')


def create_circular_mask_with_spider(resolution: int,
                                     bkg_value: int = 0) -> torch.Tensor:
    """Creates a circular mask with an inner "spider" circle removed.

    Parameters
    ----------
    resolution : int
        The resolution of the square mask.
    bkg_value : int
        The background value to set for the masked areas. Defaults to 0.

    Returns
    -------
    torch.Tensor : np.ndarray
        A 2D tensor representing the mask.
    """
    # Create a circular mask
    center = (int(resolution / 2), int(resolution / 2))
    radius = min(center)
    Y, X = np.ogrid[:resolution, :resolution]
    mask = (X - center[0])**2 + (Y - center[1])**2 > radius**2

    # Remove the inner circle (spider)
    spider_radius = int(0.22 * resolution)
    spider_mask = (X - center[0])**2 + (Y - center[1])**2 < spider_radius**2

    # Apply background value to the mask and spider mask
    final_mask = np.ones((resolution, resolution), dtype=np.uint8)
    final_mask[mask] = bkg_value
    final_mask[spider_mask] = bkg_value

    return torch.Tensor(final_mask)



"""This module provides utility functions for loading and saving model
configurations and states.

It includes functions to load configuration files, save model states,
load model states, and load the latest model state from a directory.
"""

from __future__ import annotations

import os
import sys

import torch
import yaml

from speckcn2.utils import ensure_directory


def load_config(config_file_path: str) -> dict:
    """Load the configuration file from a given path.

    This function reads a YAML configuration file and returns its contents as a dictionary.

    Parameters
    ----------
    config_file_path : str
        Path to the .yaml configuration file

    Returns
    -------
    config : dict
        Dictionary containing the configuration
    """
    with open(config_file_path, 'r') as file:
        config = yaml.safe_load(file)
    return config


def save(model: torch.nn.Module,
         datadirectory: str,
         early_stop: bool = False) -> None:
    """Save the model state and the model itself to a specified directory.

    This function saves the model's state dictionary and other relevant information
    such as epoch, loss, validation loss, and time to a file in the specified directory.

    Parameters
    ----------
    model : torch.nn.Module
        The model to save
    datadirectory : str
        The directory where the data is stored
    early_stop: bool
        If True, the model corresponds to the moment when early stop was triggered
    """
    model_state = {
        'epoch': model.epoch,
        'loss': model.loss,
        'val_loss': model.val_loss,
        'time': model.time,
        'model_state_dict': model.state_dict(),
    }

    if not early_stop:
        savename = f'{datadirectory}/{model.name}_states/{model.name}_{model.epoch[-1]}.pth'
    else:
        savename = f'{datadirectory}/{model.name}_states/{model.name}_{model.epoch[-1]}_earlystop.pth'

    torch.save(model_state, savename)


def load(model: torch.nn.Module,
         datadirectory: str,
         epoch: int,
         early_stop: bool = False) -> None:
    """Load the model state and the model itself from a specified directory and
    epoch.

    This function loads the model's state dictionary and other relevant information
    such as epoch, loss, validation loss, and time from a file in the specified directory.

    Parameters
    ----------
    model : torch.nn.Module
        The model to load
    datadirectory : str
        The directory where the data is stored
    epoch : int
        The epoch of the model
    early_stop: bool
        If True, the last state reached the early stop condition
    """
    if early_stop:
        model_state = torch.load(
            f'{datadirectory}/{model.name}_states/{model.name}_{epoch}_earlystop.pth'
        )
        model.early_stop = True
    else:
        model_state = torch.load(
            f'{datadirectory}/{model.name}_states/{model.name}_{epoch}.pth')

    model.epoch = model_state['epoch']
    model.loss = model_state['loss']
    model.val_loss = model_state['val_loss']
    model.time = model_state['time']
    model.load_state_dict(model_state['model_state_dict'], strict=False)

    assert model.epoch[
        -1] == epoch, 'The epoch of the model is not the same as the one loaded'


def load_model_state(model: torch.nn.Module,
                     datadirectory: str) -> tuple[torch.nn.Module, int]:
    """Loads the latest model state from the given directory.

    This function checks the specified directory for the latest model state file,
    loads it, and updates the model with the loaded state. If no state is found,
    it initializes the model state.
    If the training was stopped after meeting an early stop condition, this function
    signals that the training should not be continued.

    Parameters
    ----------
    model : torch.nn.Module
        The model to load the state into
    datadirectory : str
        The directory where the model states are stored

    Returns
    -------
    model : torch.nn.Module
        The model with the loaded state
    last_model_state : int
        The number of the last model state
    """
    # Print model information
    print(model)
    model.nparams = sum(p.numel() for p in model.parameters())
    print(f'\n--> Nparams = {model.nparams}')

    fulldirname = f'{datadirectory}/{model.name}_states'
    ensure_directory(fulldirname)

    # First check if there was an early stop
    earlystop = [
        filename for filename in os.listdir(fulldirname)
        if 'earlystop' in filename
    ]
    if len(earlystop) == 0:
        # If there was no early stop, check what is the last model state
        try:
            last_model_state = sorted([
                int(file_name.split('.pth')[0].split('_')[-1])
                for file_name in os.listdir(fulldirname)
            ])[-1]
        except Exception as e:
            print(f'Warning: {e}')
            last_model_state = 0

        if last_model_state > 0:
            print(
                f'Loading model at epoch {last_model_state}, from {datadirectory}'
            )
            load(model, datadirectory, last_model_state)
            return model, last_model_state
        else:
            print('No pretrained model to load')

            # Initialize some model state measures
            model.loss = []
            model.val_loss = []
            model.time = []
            model.epoch = []

            return model, 0
    elif len(earlystop) == 1:
        filename = earlystop[0]
        print(f'Loading the early stop state {filename}')
        last_model_state = int(filename.split('_')[-2])
        load(model, datadirectory, last_model_state, early_stop=True)
        return model, last_model_state
    else:
        print(
            f'Error: more than one early stop state found. This is not correct. This is the list: {earlystop}'
        )
        sys.exit(0)



"""This module implements various loss functions for training machine learning
models.

It includes custom loss functions that extend PyTorch's nn.Module,
allowing for flexible and efficient computation of loss values during
training. The loss functions handle different scenarios such as
classification, regression, and segmentation tasks. They incorporate
techniques like weighted losses, focal losses, and smooth L1 losses to
address class imbalances and improve model performance. The module
ensures that the loss calculations are compatible with PyTorch's
autograd system, enabling seamless integration into training loops.
"""
from __future__ import annotations

from typing import Callable

import numpy as np
import torch
import torch.nn as nn

from .normalizer import Normalizer


class ComposableLoss(nn.Module):
    """Compose the loss function using several terms. The importance of each
    term has to be specified in the configuration file. Each term with a >0
    weight will be added to the loss function.

    The loss term available are:
    - MSE: mean squared error between predicted and target normalized screen tags
    - MAE: mean absolute error between predicted and target normalized screen tags
    - JMSE: mean squared error between predicted and target J
    - JMAE: mean absolute error between predicted and target J
    - Pearson: Pearson correlation coefficient between predicted and target J
    - Fried: Fried parameter r0
    - Isoplanatic: Isoplanatic angle theta0
    - Rytov: Rytov variance sigma_r^2 that will be computed on log averaged Cn2
    - Scintillation_w: scintillation index for weak turbulence
    - Scintillation_m: scintillation index for moderate-strong turbulence

    Parameters
    ----------
    config : dict
        Dictionary containing the configuration
    nz : Normalizer
        Normalizer object to be used to extract J in its original scale
    device : torch.device
        The device to use for the computation
    validation : bool
        If true, use the validation parameters from config
    """

    def __init__(self,
                 config: dict,
                 nz: Normalizer,
                 device: torch.device,
                 validation: bool = False):
        super(ComposableLoss, self).__init__()
        if validation:
            if 'val_loss' in config:
                config['loss'] = config['val_loss']
            else:
                print('[!] Warning: Validation loss not found in config.yaml,',
                      'keeping track of training loss instead')
        self.device = device
        self.loss_functions: dict[str, Callable] = {
            'MSE': torch.nn.MSELoss(),
            'MAE': torch.nn.L1Loss(),
            'JMSE': self._MSELoss,
            'JMAE': self._L1Loss,
            'Cn2MSE': self._do_nothing,
            'Cn2MAE': self._do_nothing,
            'Pearson': self._PearsonCorrelationLoss,
            'Fried': self._FriedLoss,
            'Isoplanatic': self._IsoplanaticLoss,
            'Rytov': self._RytovLoss,
            'Scintillation_w': self._ScintillationWeakLoss,
            'Scintillation_ms': self._ScintillationModerateStrongLoss,
        }
        self.loss_weights = {
            loss_name: config['loss'].get(loss_name, 0)
            for loss_name in self.loss_functions.keys()
        }
        self.total_weight = sum(self.loss_weights.values())
        self._select_loss_needed()

        # And get some useful parameters for the loss functions
        # the parameters are explained at:
        # https://males-project.github.io/SpeckleCn2Profiler/example/#configuration-file-explanation
        self.h = torch.Tensor([float(x) for x in config['speckle']['hArray']])
        self.k = 2 * torch.pi / (config['speckle'].get('lambda', 550) * 1e-9)
        self.cosz = np.cos(np.deg2rad(config['speckle'].get('z', 0)))
        self.secz = 1 / self.cosz
        self.p_fr = 0.423 * self.k**2 * self.secz
        self.p_iso = self.cosz**(8. / 5.) / ((2.91 * self.k**2)**(3. / 5.))
        self.p_scw = 2.25 * self.k**(7. / 6.) * self.secz**(11. / 6.)

        # We need to ba able to recover the tags
        self.recover_tag = nz.recover_tag
        # Move tensors to the device
        self.h = self.h.to(self.device)

    def _select_loss_needed(self):
        """Select the loss functions that will be used in the loss calculation.

        This allows to skip the computation of optional elements like
        Cn2.
        """
        self.loss_needed = {
            loss_name: loss_fn
            for loss_name, loss_fn in self.loss_functions.items()
            if self.loss_weights[loss_name] > 0
        }
        self.Cn2required = any([
            loss_name in [
                'Cn2MSE', 'Cn2MAE', 'Fried', 'Isoplanatic', 'Rytov',
                'Scintillation_w', 'Scintillations_ms'
            ] for loss_name in self.loss_needed.keys()
        ])

    def forward(self, pred: torch.Tensor,
                target: torch.Tensor) -> tuple[torch.Tensor, dict]:
        """Forward pass of the loss function.

        Parameters
        ----------
        pred : torch.Tensor
            The predicted screen tags
        target : torch.Tensor
            The target screen tags

        Returns
        -------
        loss : torch.Tensor
            The composed loss
        losses : dict
            Dictionary containing the individual losses
        """
        total_loss = 0
        losses = {}

        if self.Cn2required:
            Cn2_pred = self.reconstruct_cn2(pred)
            Cn2_target = self.reconstruct_cn2(target)

        for loss_name, loss_fn in self.loss_needed.items():
            weight = self.loss_weights[loss_name]
            if loss_name in ['MAE', 'MSE']:
                this_loss = loss_fn(pred, target)
            else:
                this_loss = loss_fn(pred, target, Cn2_pred, Cn2_target)
            total_loss += weight * this_loss
            losses[loss_name] = this_loss

        return total_loss / self.total_weight, losses

    def _do_nothing(*args, **kwargs):
        pass

    def get_J(self, Jnorm: torch.Tensor) -> torch.Tensor:
        """Recover J from the normalized tags. This needs to be done to compute
        Cn2.

        Parameters
        ----------
        Jnorm : torch.Tensor
            The normalized screen tags between 0 and 1

        Returns
        -------
        J : torch.Tensor
            The recovered screen tags
        """

        if Jnorm.ndim == 1:
            Jnorm = Jnorm[None, :]

        J = []
        for i in range(Jnorm.shape[0]):
            J.append(
                torch.tensor([
                    10**self.recover_tag[j](Jnorm[i][j], i)
                    for j in range(len(Jnorm[i]))
                ],
                             requires_grad=True).to(Jnorm.device))
        J = torch.stack(J)
        return J

    def reconstruct_cn2(self, Jnorm: torch.Tensor) -> torch.Tensor:
        """ Reconstruct Cn2 from screen tags
        c_i = J_i / (h[i+1] - h[i])

        Parameters
        ----------
        Jnorm : torch.Tensor
            The screen tags normalized between 0 and 1

        Returns
        -------
        Cn2 : torch.Tensor
            The Cn2 reconstructed from the screen tags, assuming a uniform profile
        """
        J = self.get_J(Jnorm)
        Cn2 = J / (self.h[1:] - self.h[:-1])
        return Cn2

    def _MSELoss(self,
                 pred: torch.Tensor,
                 target: torch.Tensor,
                 Cn2p: torch.Tensor = None,
                 Cn2t: torch.Tensor = None) -> torch.Tensor:
        """Mean squared error loss function.

        Parameters
        ----------
        pred : torch.Tensor
            The predicted screen tags
        target : torch.Tensor
            The target screen tags
        Cn2p : torch.Tensor
            The predicted Cn2
        Cn2t : torch.Tensor
            The target Cn2

        Returns
        -------
        loss : torch.Tensor
            The mean squared error loss
        """
        loss = self.loss_weights['JMSE'] * torch.mean(
            (pred - target)**2 / ((target**2).sum(-1).unsqueeze(-1) + 1e-5))
        # Optionally add the Cn2MSE loss
        if self.loss_weights['Cn2MSE'] > 0:
            loss += self.loss_weights['Cn2MSE'] * torch.mean(
                (Cn2p - Cn2t)**2 / ((Cn2t**2).sum(-1).unsqueeze(-1) + 1e-5))
        return loss

    def _L1Loss(self,
                pred: torch.Tensor,
                target: torch.Tensor,
                Cn2p: torch.Tensor = None,
                Cn2t: torch.Tensor = None) -> torch.Tensor:
        """Mean absolute error loss function.

        Parameters
        ----------
        pred : torch.Tensor
            The predicted screen tags
        target : torch.Tensor
            The target screen tags
        Cn2p : torch.Tensor
            The predicted Cn2
        Cn2t : torch.Tensor
            The target Cn2

        Returns
        -------
        loss : torch.Tensor
            The mean absolute error loss
        """
        loss = self.loss_weights['JMAE'] * torch.mean(
            torch.abs(pred - target) / (target.sum(-1).unsqueeze(-1) + 1e-5))
        # Optionally add the Cn2MAE loss
        if self.loss_weights['Cn2MAE'] > 0:
            loss += self.loss_weights['Cn2MAE'] * torch.mean(
                torch.abs(Cn2p - Cn2t) / (Cn2t.sum(-1).unsqueeze(-1) + 1e-5))
        return loss

    def _PearsonCorrelationLoss(self, pred: torch.Tensor, target: torch.Tensor,
                                Cn2p: torch.tensor,
                                Cn2t: torch.Tensor) -> torch.Tensor:
        """Pearson correlation coefficient loss function.

        Parameters
        ----------
        pred : torch.Tensor
            The predicted screen tags
        target : torch.Tensor
            The target screen tags
        Cn2p : torch.Tensor
            The predicted Cn2
        Cn2t : torch.Tensor
            The target Cn2

        Returns
        -------
        loss : torch.Tensor
            The Pearson correlation coefficient loss
        """

        mean_pred = torch.mean(pred)
        mean_target = torch.mean(target)
        # Calculate covariance
        cov_xy = torch.mean((pred - mean_pred) * (target - mean_target))
        # Calculate standard deviations
        std_x = torch.std(pred)
        std_y = torch.std(target)
        # Calculate Pearson correlation coefficient
        corr = cov_xy / (std_x * std_y + 1e-5)
        # The loss is 1 - correlation to be minimized
        loss = 1.0 - corr

        return loss

    def get_FriedParameter(self, Jnorm: torch.Tensor) -> torch.Tensor:
        """Compute the Fried parameter r0 from the screen tags."""
        J = torch.Tensor(self.get_J(Jnorm))
        return (self.p_fr * torch.sum(J))**(-3 / 5)

    def _FriedLoss(self, pred: torch.Tensor, target: torch.Tensor,
                   Cn2p: torch.Tensor, Cn2t: torch.Tensor) -> torch.Tensor:
        """Fried parameter r0 loss function. The difference between real value
        and prediction is normalized by the real value, since small values are
        more important to get correctly.

        Parameters
        ----------
        pred : torch.Tensor
            The predicted screen tags
        target : torch.Tensor
            The target screen tags
        Cn2p : torch.Tensor
            The predicted Cn2
        Cn2t : torch.Tensor
            The target Cn2

        Returns
        -------
        loss : torch.Tensor
            The Fried parameter r0 loss
        """
        r0p = self.get_FriedParameter(pred)
        r0t = self.get_FriedParameter(target)
        loss = torch.abs(r0p - r0t) / r0t

        return loss

    def get_IsoplanaticAngle(self, Cn2: torch.Tensor) -> torch.Tensor:
        """Compute the isoplanatic angle theta0 from the screen tags."""
        # Integrate Cn2*z^(5/3)
        integral = torch.sum(
            Cn2 * (self.h[1:]**(8 / 3) - self.h[:-1]**(8 / 3))) * 3 / 8
        # Then I can compute theta0
        return self.p_iso / (integral**(3 / 5))

    def _IsoplanaticLoss(self, pred: torch.Tensor, target: torch.Tensor,
                         Cn2p: torch.Tensor,
                         Cn2t: torch.Tensor) -> torch.Tensor:
        """Isoplanatic angle theta0 loss function. The loss is normalized by
        the real value, since small values are more important to get correctly.

        Parameters
        ----------
        pred : torch.Tensor
            The predicted screen tags
        target : torch.Tensor
            The target screen tags
        Cn2p : torch.Tensor
            The predicted Cn2
        Cn2t : torch.Tensor
            The target Cn2

        Returns
        -------
        loss : torch.Tensor
            The isoplanatic angle theta0 loss
        """
        isp = self.get_IsoplanaticAngle(Cn2p)
        ist = self.get_IsoplanaticAngle(Cn2t)
        loss = torch.abs(isp - ist) / ist

        return loss

    def _RytovLoss(self, pred: torch.Tensor, target: torch.Tensor,
                   Cn2p: torch.Tensor, Cn2t: torch.Tensor) -> torch.Tensor:
        """Rytov variance sigma_r^2 loss function."""
        # //TODO: discuss about this. Since it is a variance measure,
        # it would have to be compared not to a single target,
        # but to the average of the dataset

        # throw a not implemented yet warning
        print('RytovLoss not implemented yet. Use scintillation index instead')
        return 0

    def get_ScintillationWeak(self, Cn: torch.Tensor) -> torch.Tensor:
        """Compute the scintillation index for weak turbulence sigma^2 from the
        screen tags."""
        # Integrate Cn2*z^(5/6)
        integral = torch.sum(
            Cn * (self.h[1:]**(11 / 6) - self.h[:-1]**(11 / 6))) * 6 / 11
        # Then I can compute sigma^2
        return self.p_scw * integral

    def _ScintillationWeakLoss(self, pred: torch.Tensor, target: torch.Tensor,
                               Cn2p: torch.Tensor,
                               Cn2t: torch.Tensor) -> torch.Tensor:
        """Scintillation index for weak turbulence loss function. The loss is
        normalized by the real value, since small values are more important to
        get correctly.

        Parameters
        ----------
        pred : torch.Tensor
            The predicted screen tags
        target : torch.Tensor
            The target screen tags
        Cn2p : torch.Tensor
            The predicted Cn2
        Cn2t : torch.Tensor
            The target Cn2

        Returns
        -------
        loss : torch.Tensor
            The scintillation index for weak turbulence loss
        """
        swp = self.get_ScintillationWeak(Cn2p)
        swt = self.get_ScintillationWeak(Cn2t)
        loss = torch.abs(swp - swt) / swt

        return loss

    def get_ScintillationModerateStrong(self, x: torch.Tensor) -> torch.Tensor:
        """Compute the scintillation index for moderate-strong turbulence
        sigma^2 from the screen tags."""
        wsigma2 = self.get_ScintillationWeak(x)
        return torch.exp(wsigma2 * 0.49 / (1 + 1.11 * wsigma2**(6 / 5)) +
                         0.51 * wsigma2 / (1 + 0.69 * wsigma2**(6 / 5)))

    def _ScintillationModerateStrongLoss(self, pred: torch.Tensor,
                                         target: torch.Tensor,
                                         Cn2p: torch.Tensor,
                                         Cn2t: torch.Tensor) -> torch.Tensor:
        """Scintillation index for moderate-strong turbulence loss function.

        Parameters
        ----------
        pred : torch.Tensor
            The predicted screen tags
        target : torch.Tensor
            The target screen tags
        Cn2p : torch.Tensor
            The predicted Cn2
        Cn2t : torch.Tensor
            The target Cn2

        Returns
        -------
        loss : torch.Tensor
            The scintillation index for moderate-strong turbulence loss
        """
        smp = self.get_ScintillationModerateStrong(Cn2p)
        smt = self.get_ScintillationModerateStrong(Cn2t)
        loss = torch.mean(torch.abs(smp - smt) / smt)

        return loss

    def _get_all_measures(self, pred: torch.Tensor, target: torch.Tensor,
                          Cn2p: torch.Tensor, Cn2t: torch.Tensor) -> dict:
        """Get all the measures available.

        Parameters
        ----------
        pred : torch.Tensor
            The predicted screen tags
        target : torch.Tensor
            The target screen tags
        Cn2p : torch.Tensor
            The predicted Cn2
        Cn2t : torch.Tensor
            The target Cn2

        Returns
        -------
        measures : dict
            Dictionary containing the measures
        """
        measures = {}
        measures['Fried_true'] = self.get_FriedParameter(target)
        measures['Fried_pred'] = self.get_FriedParameter(pred)
        measures['Isoplanatic_true'] = self.get_IsoplanaticAngle(Cn2t)
        measures['Isoplanatic_pred'] = self.get_IsoplanaticAngle(Cn2p)
        measures['Scintillation_w_true'] = self.get_ScintillationWeak(Cn2t)
        measures['Scintillation_w_pred'] = self.get_ScintillationWeak(Cn2p)

        return measures



# Import any modules or subpackages here
from __future__ import annotations

from .io import load, load_config, load_model_state, save
from .loss import ComposableLoss
from .mlmodels import EnsembleModel, get_a_resnet, setup_model
from .mlops import score, train
from .normalizer import Normalizer
from .plots import (
    plot_histo_losses,
    plot_J_error_details,
    plot_loss,
    plot_param_histo,
    plot_param_vs_loss,
    plot_samples_in_ensemble,
    plot_time,
)
from .postprocess import average_speckle_input, average_speckle_output, screen_errors, tags_distribution
from .preprocess import prepare_data, train_test_split
from .utils import setup_optimizer

__version__ = '1.0.1'
__all__ = [
    'load_config',
    'prepare_data',
    'Normalizer',
    'train_test_split',
    'setup_model',
    'get_a_resnet',
    'EnsembleModel',
    'ComposableLoss',
    'setup_optimizer',
    'train',
    'score',
    'tags_distribution',
    'screen_errors',
    'average_speckle_input',
    'average_speckle_output',
    'plot_J_error_details',
    'plot_loss',
    'plot_time',
    'save',
    'load',
    'load_model_state',
    'plot_histo_losses',
    'plot_param_vs_loss',
    'plot_param_histo',
    'plot_samples_in_ensemble',
]



"""This module contains functions for training and evaluating a neural network
model using PyTorch. It includes the following key components:

1. `train`: Trains the model for a specified number of epochs, logs training
   and validation losses, and saves the model state at specified intervals.
2. `score`: Evaluates the model on a test dataset, calculates various metrics,
   and generates plots for a specified number of test samples.

The module relies on several external utilities and models from the `speckcn2`
package, including `EnsembleModel`, `ComposableLoss`, and `Normalizer`.
"""

from __future__ import annotations

import os
import pickle
import random
import time

import h5py
import numpy as np
import torch
import torchvision.transforms as transforms
from PIL import Image

from speckcn2.normalizer import Normalizer
from speckcn2.transformations import PolarCoordinateTransform, ShiftRowsTransform, ToUnboundTensor
from speckcn2.utils import ensure_directory, plot_preprocessed_image


def assemble_transform(conf: dict) -> transforms.Compose:
    """Assembles the transformation to apply to each image.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration

    Returns
    -------
    transform : torchvision.transforms.Compose
        Transformation to apply to the images
    """
    list_transforms = []

    if conf['preproc']['centercrop'] > 0:
        # Take only the center of the image
        list_transforms.append(
            transforms.CenterCrop(conf['preproc']['centercrop']))

    if conf['preproc']['polarize']:
        # make the image larger for better polar conversion
        list_transforms.append(transforms.Resize(conf['preproc']['polresize']))
        # Convert to polar coordinates
        list_transforms.append(PolarCoordinateTransform())

    if conf['preproc']['randomrotate'] and not conf['preproc']['polarize']:
        # Randomly rotate the image, since it is symmetric (not if it is polarized)
        list_transforms.append(transforms.RandomRotation(degrees=(-180, 180)))

    if conf['preproc']['resize']:
        # Optionally, downscale it
        list_transforms.append(
            transforms.Resize(
                (conf['preproc']['resize'], conf['preproc']['resize'])))

    if conf['preproc']['equivariant'] and conf['preproc']['polarize']:
        # Apply the equivariant transform, which makes sense only in polar coordinates
        list_transforms.append(ShiftRowsTransform())

    list_transforms.append(ToUnboundTensor())

    return transforms.Compose(list_transforms)


def prepare_data(
    conf: dict,
    nimg_print: int = 5,
) -> tuple[list, list, list]:
    """If not already available, preprocesses the data by loading images and
    tags from the given directory, applying a transformation to the images.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    nimg_print: int
        Number of images to print

    Returns
    -------
    all_images : list
        List of all images
    all_tags : list
        List of all tags
    all_ensemble_ids : list
        List of all ensemble ids, representing images from the same Cn2 profile
    """
    datadirectory = conf['speckle']['datadirectory']
    dataname = conf['preproc']['dataname']
    tagname = dataname.replace('images', 'tags')
    ensemblename = dataname.replace('images', 'ensemble')

    # First, check if the data has already been preprocessed
    if os.path.exists(os.path.join(datadirectory, dataname)):
        print(f'*** Loading preprocessed data from {dataname}')
        # If so, load it
        all_images = torch.load(os.path.join(datadirectory, dataname))
        all_tags = torch.load(os.path.join(datadirectory, tagname))
        all_ensemble_ids = torch.load(os.path.join(datadirectory,
                                                   ensemblename))
        # print the info about the dataset
        print(f'*** There are {len(all_images)} images in the dataset.')
    else:
        # Check if there is at least one image file in the directory
        if not any('.h5' in file_name
                   for file_name in os.listdir(datadirectory)):
            raise FileNotFoundError(
                'No image files found in the directory. Please provide the '
                'correct path to the data directory.')
        # Otherwise, preprocess the raw data separating the single images
        all_images, all_tags, all_ensemble_ids = imgs_as_single_datapoint(
            conf, nimg_print)

    # Get the average value of the pixels, excluding the 0 values
    non_zero_pixels = 0
    sum_pixels = 0
    for image in all_images:
        non_zero_pixels_in_image = image[image != 0]
        non_zero_pixels += non_zero_pixels_in_image.numel()
        sum_pixels += torch.sum(non_zero_pixels_in_image)
    pixel_average = sum_pixels / non_zero_pixels
    print('*** Pixel average:', pixel_average)
    # and store it in the config
    conf['preproc']['pixel_average'] = pixel_average

    return all_images, all_tags, all_ensemble_ids


def imgs_as_single_datapoint(
    conf: dict,
    nimg_print: int = 5,
) -> tuple[list, list, list]:
    """Preprocesses the data by loading images and tags from the given
    directory, applying a transformation to the images. Each image is treated
    as a single data point.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    nimg_print: int
        Number of images to print

    Returns
    -------
    all_images : list
        List of all images
    all_tags : list
        List of all tags
    all_ensemble_ids : list
        List of all ensemble ids, representing images from the same Cn2 profile
    """
    time_start = time.time()
    datadirectory = conf['speckle']['datadirectory']
    mname = conf['model']['name']
    dataname = conf['preproc']['dataname']
    tagname = dataname.replace('images', 'tags')
    ensemblename = dataname.replace('images', 'ensemble')
    nreps = conf['preproc']['speckreps']

    # Dummy transformation to get the original image
    transform_orig = transforms.Compose([
        transforms.ToTensor(),
    ])

    # and get the transformation to apply to each image
    transform = assemble_transform(conf)

    # Get the list of images
    file_list = [
        file_name for file_name in os.listdir(datadirectory)
        if '.h5' in file_name and 'tag' not in file_name
    ]
    np.random.shuffle(file_list)
    # Optionally, do data augmentation by using each file multiple times
    # (It makes sense only in combination with random rotations)
    file_list = file_list * nreps

    all_images, all_tags, all_ensemble_ids = [], [], []
    show_image = nimg_print > 0
    if show_image:
        ensure_directory(f'{datadirectory}/imgs_to_{mname}')

    tag_files = get_tag_files(file_list, datadirectory)
    ensemble_dict = get_ensemble_dict(tag_files)

    # Load each text file as an image
    for counter, file_name in enumerate(file_list):
        # Process only if tags available
        if file_name in tag_files:

            # Construct the full path to the file
            file_path = os.path.join(datadirectory, file_name)

            # Open the HDF5 file
            print(file_path, flush=True)
            with h5py.File(file_path, 'r') as f:
                # Load the data from the 'data' dataset
                pixel_values = np.float32(f['data'][:])
                # and replace nans with 0
                np.nan_to_num(pixel_values, copy=False)

            # Create the image
            image_orig = Image.fromarray(pixel_values, mode='F')

            # Apply the transformation
            image = transform(image_orig)
            if show_image:
                image_orig = transform_orig(image_orig)

            # and add the img to the collection
            all_images.append(image)

            # Load the tags
            with h5py.File(tag_files[file_name], 'r') as f:
                if 'sample' in file_name:
                    tags = f['J'][:].reshape(8, 1)
                else:
                    tags = f['data'][:]

            # Plot the image using maplotlib
            if counter > nimg_print:
                show_image = False
            if show_image:
                plot_preprocessed_image(image_orig, image, tags, counter,
                                        datadirectory, mname, file_name)

            # Preprocess the tags
            np.log10(tags, out=tags)
            # Add the tag to the collection
            all_tags.append(tags.squeeze())

            # Get the ensemble ID
            ensemble_id = ensemble_dict[file_name]
            # and add it to the collection
            all_ensemble_ids.append(ensemble_id)

    # Finally, store them before returning
    torch.save(all_images, os.path.join(datadirectory, dataname))
    torch.save(all_tags, os.path.join(datadirectory, tagname))
    torch.save(all_ensemble_ids, os.path.join(datadirectory, ensemblename))

    print('*** Preprocessing complete.', flush=True)
    print('It took',
          time.time() - time_start, 'seconds to preprocess the data.')

    return all_images, all_tags, all_ensemble_ids


def get_tag_files(file_list: list, datadirectory: str) -> dict:
    """Function to check the existence of tag files for each image file.

    Parameters
    ----------
    file_list : list
        List of image files
    datadirectory : str
        The directory containing the data

    Returns
    -------
    tag_files : dict
        Dictionary of image files and their corresponding tag files
    """
    tag_files = {}
    for file_name in file_list:
        if 'MALES' in file_name:
            ftagname = file_name.replace('.h5', '_tag.h5')
        elif 'sample' in file_name:
            ftagname = file_name.rpartition('-')[0] + '_tag.h5'
        else:
            ftagname = file_name.rpartition('_')[0] + '_tag.h5'

        tag_path = os.path.join(datadirectory, ftagname)
        if os.path.exists(tag_path):
            tag_files[file_name] = tag_path
        else:
            print(f'*** Warning: tag file {ftagname} not found.')
    return tag_files


def get_ensemble_dict(tag_files: dict) -> dict:
    """Function to associate each Cn2 profile to an ensemble ID for parallel
    processing.

    Parameters
    ----------
    tag_files : dict
        Dictionary of image files and their corresponding tag files

    Returns
    -------
    ensemble_dict : dict
        Dictionary of image files and their corresponding ensemble IDs
    """
    ensembles = {}
    ensemble_counter = 1
    for value in tag_files.values():
        # Check if the value is already assigned an ID
        if value not in ensembles:
            # Assign a new ID if it's a new value
            ensembles[value] = ensemble_counter
            ensemble_counter += 1
    return {key: ensembles[value] for key, value in tag_files.items()}


def train_test_split(
    all_images: list[torch.tensor],
    all_tags: list[np.ndarray],
    all_ensemble_ids: list[int],
    nz: Normalizer,
) -> tuple[list, list]:
    """Splits the data into training and testing sets.

    Parameters
    ----------
    all_images : list
        List of images
    all_tags : list
        List of tags
    all_ensemble_ids : list
        List of ensemble ids
    nz: Normalizer
        The normalizer object to preprocess the data

    Returns
    -------
    train_set : list
        Training dataset
    test_set : list
        Testing dataset
    """
    # Get the config dict
    config = nz.conf
    # extract the model parameters
    modelname = config['model']['name']
    datadirectory = config['speckle']['datadirectory']
    ttsplit = config['hyppar'].get('ttsplit', 0.8)
    ensemble_size = config['preproc'].get('ensemble', 1)
    average_size = config['preproc'].get('average', 0)

    # Check if the training and test set are already prepared
    train_file = f'{datadirectory}/train_set_{modelname}.pickle'
    test_file = f'{datadirectory}/test_set_{modelname}.pickle'

    if os.path.isfile(train_file) and os.path.isfile(test_file):
        print('Loading the training and testing set...', flush=True)
        nz._normalizing_functions(all_images, all_tags, all_ensemble_ids)
        train_data = pickle.load(open(train_file, 'rb'))
        test_data = pickle.load(open(test_file, 'rb'))
        print(f'*** There are {len(train_data)} images in the training set, ')
        print(f'*** and {len(test_data)} images in the testing set.')
        return train_data, test_data

    # If the data are not already prepared, first I normalize them using the Normalizer object
    print('Normalizing the images and tags...', flush=True)
    dataset = nz.normalize_imgs_and_tags(all_images, all_tags,
                                         all_ensemble_ids)

    if average_size > 1 and ensemble_size > 1:
        raise ValueError(
            'The average_size and ensemble_size cannot be set at the same time.'
        )
    elif average_size > 1:
        dataset = create_average_dataset(dataset, average_size)
        print_average_info(dataset, average_size, ttsplit)
    elif ensemble_size > 1:
        dataset = create_ensemble_dataset(dataset, ensemble_size)
        print_ensemble_info(dataset, ensemble_size, ttsplit)
    else:
        print_dataset_info(dataset, ttsplit)

    train_set, test_set = split_dataset(dataset, ttsplit)

    pickle.dump(train_set, open(train_file, 'wb'))
    pickle.dump(test_set, open(test_file, 'wb'))

    return train_set, test_set


def create_average_dataset(dataset: list, average_size: int) -> list:
    """Creates a dataset of averages from a dataset of single images. The
    averages are created by grouping together average_size images.

    Parameters
    ----------
    dataset : list
        List of single images
    average_size : int
        The number of images that will be averaged together

    Returns
    -------
    average_dataset : list
        List of averages
    """
    split_averages: dict = {}
    for item in dataset:
        key = item[-1]
        split_averages.setdefault(key, []).append(item)

    average_dataset: list = []
    for average in split_averages.values():
        # * In each average, take n_groups groups of average_size datapoints
        n_groups = len(average) // average_size
        if n_groups < 1:
            raise ValueError(f'Average size {average_size} is too large '
                             f'for groups with size {len(average)}')
        # Extract the averages randomly
        sample = random.sample(average, n_groups * average_size)
        # Split the sample into groups of average_size
        list_to_avg = [
            sample[i:i + average_size]
            for i in range(0, n_groups * average_size, average_size)
        ]
        # Average the groups
        averages = [
            tuple(
                sum(element) / average_size
                for i, element in enumerate(zip(*group)))
            for group in list_to_avg
        ]
        average_dataset.extend(averages)

    random.shuffle(average_dataset)
    return average_dataset


def print_average_info(dataset: list, average_size: int, ttsplit: int):
    """Prints the information about the average dataset.

    Parameters
    ----------
    dataset : list
        The average dataset
    average_size : int
        The number of images in each average
    ttsplit : int
        The train-test split
    """
    train_size = int(ttsplit * len(dataset))
    print(
        f'*** There are {len(dataset)} average groups in the dataset, '
        f'that I split in {train_size} for training and '
        f'{len(dataset) - train_size} for testing. Each average is composed by '
        f'{average_size} images. This corresponds to {train_size * average_size} '
        f'for training and {(len(dataset) - train_size) * average_size} for testing.'
    )


def create_ensemble_dataset(dataset: list, ensemble_size: int) -> list:
    """Creates a dataset of ensembles from a dataset of single images. The
    ensembles are created by grouping together ensemble_size images. These
    images will be used to train the model in parallel.

    Parameters
    ----------
    dataset : list
        List of single images
    ensemble_size : int
        The number of images that will be processed together as an ensemble

    Returns
    -------
    ensemble_dataset : list
        List of ensembles
    """
    split_ensembles: dict = {}
    for item in dataset:
        key = item[-1]
        split_ensembles.setdefault(key, []).append(item)

    ensemble_dataset: list = []
    for ensemble in split_ensembles.values():
        # * In each ensemble, take n_groups groups of ensemble_size datapoints
        n_groups = len(ensemble) // ensemble_size
        if n_groups < 1:
            raise ValueError(f'Ensemble size {ensemble_size} is too large '
                             f'for ensembles with size {len(ensemble)}')
        # Extract the ensembles randomly
        sample = random.sample(ensemble, n_groups * ensemble_size)
        # Split the sample into groups of ensemble_size
        ensemble_dataset.extend(sample[i:i + ensemble_size]
                                for i in range(0, n_groups *
                                               ensemble_size, ensemble_size))

    random.shuffle(ensemble_dataset)
    return ensemble_dataset


def print_ensemble_info(dataset: list, ensemble_size: int, ttsplit: int):
    """Prints the information about the ensemble dataset.

    Parameters
    ----------
    dataset : list
        The ensemble dataset
    ensemble_size : int
        The number of images in each ensemble
    ttsplit : int
        The train-test split
    """

    train_size = int(ttsplit * len(dataset))
    print(
        f'*** There are {len(dataset)} ensemble groups in the dataset, '
        f'that I split in {train_size} for training and '
        f'{len(dataset) - train_size} for testing. Each ensemble is composed by '
        f'{ensemble_size} images. This corresponds to {train_size * ensemble_size} '
        f'for training and {(len(dataset) - train_size) * ensemble_size} for testing.'
    )


def print_dataset_info(dataset: list, ttsplit: int):
    """Prints the information about the dataset.

    Parameters
    ----------
    dataset : list
        The dataset
    ttsplit : int
        The train-test split
    """
    train_size = int(ttsplit * len(dataset))
    print(f'*** There are {len(dataset)} images in the dataset, '
          f'{train_size} for training and '
          f'{len(dataset) - train_size} for testing.')


def split_dataset(dataset: list, ttsplit: int) -> tuple[list, list]:
    """Splits the dataset into training and testing sets.

    Parameters
    ----------
    dataset : list
        The dataset
    ttsplit : int
        The train-test split

    Returns
    -------
    train_set : list
        The training set
    test_set : list
        The testing set
    """
    # First shuffle the dataset
    random.shuffle(dataset)
    train_size = int(ttsplit * len(dataset))
    return dataset[:train_size], dataset[train_size:]



"""This module contains the definition of the EnsembleModel class and a
setup_model function.

The EnsembleModel class is a wrapper that allows any model to be used
for ensembled data. The setup_model function initializes and returns a
model based on the provided configuration.
"""

from __future__ import annotations

import itertools
import random

import numpy as np
import torch
import torchvision
from torch import nn

from speckcn2.io import load_model_state
from speckcn2.scnn import SteerableCNN, create_final_block


class EnsembleModel(nn.Module):
    """Wrapper that allows any model to be used for ensembled data."""

    def __init__(self, conf: dict, device: torch.device):
        """Initializes the EnsembleModel.

        Parameters
        ----------
        conf: dict
            The global configuration containing the model parameters.
        device : torch.device
            The device to use
        """
        super(EnsembleModel, self).__init__()

        self.ensemble_size = conf['preproc'].get('ensemble', 1)
        self.device = device
        self.uniform_ensemble = conf['preproc'].get('ensemble_unif', False)
        resolution = conf['preproc']['resize']
        self.D = conf['noise']['D']
        self.t = conf['noise']['t']
        self.snr = conf['noise']['snr']
        self.dT = conf['noise']['dT']
        self.dO = conf['noise']['dO']
        self.rn = conf['noise']['rn']
        self.fw = conf['noise']['fw']
        self.bit = conf['noise']['bit']
        self.discretize = conf['noise']['discretize']
        self.rot_sym = conf['noise'].get('rotation_sym', 0)
        if self.rot_sym > 0:
            self.rot_fold = 360 // self.rot_sym
        self.apply_masks = conf['noise'].get('apply_masks', False)
        if self.apply_masks:
            self.mask_D, self.mask_d, self.mask_X, self.mask_Y = self.create_masks(
                resolution)

    def forward(self, model, batch_ensemble):
        """Forward pass through the model.

        Parameters
        ----------
        model : torch.nn.Module
            The model to use
        batch_ensemble : list
            Each element is a batch of an ensemble of samples.
        """

        if self.ensemble_size == 1:
            batch = batch_ensemble
            # If no ensembling, each element of the batch is a tuple (image, tag, ensemble_id)
            images, tags, ensembles = zip(*batch)
            images = torch.stack(images).to(self.device)
            images = self.apply_noise(images)
            tags = torch.tensor(np.stack(tags)).to(self.device)

            return model(images), tags, images
        else:
            batch = list(itertools.chain(*batch_ensemble))
            # Like the ensemble=1 case, I can process independently each element of the batch
            images, tags, ensembles = zip(*batch)
            images = torch.stack(images).to(self.device)
            images = self.apply_noise(images)
            tags = torch.tensor(np.stack(tags)).to(self.device)

            model_output = model(images)

            # To average the self.ensemble_size outputs of the model I extract the confidence weights
            predictions = model_output[:, :-1]
            weights = model_output[:, -1]
            if self.uniform_ensemble:
                weights = torch.ones_like(weights)
            # multiply the prediction by the weights
            weighted_predictions = predictions * weights.unsqueeze(-1)
            # and sum over the ensembles
            weighted_predictions = weighted_predictions.view(
                model_output.size(0) // self.ensemble_size, self.ensemble_size,
                -1).sum(dim=1)
            # then normalize by the sum of the weights
            sum_weights = weights.view(
                weights.size(0) // self.ensemble_size,
                self.ensemble_size).sum(dim=1)
            ensemble_output = weighted_predictions / sum_weights.unsqueeze(-1)

            # and get the tags and ensemble_id of the first element of the ensemble
            tags = tags[::self.ensemble_size]
            ensembles = ensembles[::self.ensemble_size]

            return ensemble_output, tags, images

    def apply_noise(self, image_tensor: torch.Tensor) -> torch.Tensor:
        """Processes a tensor of 2D images.

        Parameters
        ----------
        image_tensor : torch.Tensor
            Tensor of 2D images with shape (batch, channels, width, height).

        Returns
        -------
        processed_tensor : torch.Tensor
            Tensor of processed 2D images.
        """
        batch, channels, height, width = image_tensor.shape
        processed_tensor = torch.zeros_like(image_tensor)

        # Apply rotation symmetry
        if self.rot_sym > 0:
            angle = random.randint(0, self.rot_fold)
            image_tensor = torch.rot90(image_tensor, angle, (2, 3))

        # Normalize wrt optical power
        image_tensor = image_tensor / torch.mean(
            image_tensor, dim=(2, 3), keepdim=True)

        amp = self.rn * 10**(self.snr / 20)

        for i in range(batch):
            for j in range(channels):
                B = image_tensor[i, j]

                ## Apply masks
                if self.apply_masks:
                    B[self.mask_D] = 0
                    B[self.mask_d] = 0
                    B[self.mask_X] = 0
                    B[self.mask_Y] = 0

                # Add noise sources
                A = self.rn + self.rn * torch.randn(
                    height, width, device=self.device) + amp * B + torch.sqrt(
                        amp * B) * torch.randn(
                            height, width, device=self.device)

                # Make a discretized version
                if self.discretize == 'on':
                    C = torch.round(A / self.fw * 2**self.bit)
                    C[A > self.fw] = self.fw
                    C[A < 0] = 0
                else:
                    C = A

                processed_tensor[i, j] = C

        return processed_tensor

    def create_masks(
        self, resolution: int
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """Creates the masks for the circular aperture and the spider.

        Parameters
        ----------
        resolution : int
            Resolution of the images.

        Returns
        -------
        mask_D : torch.Tensor
            Mask for the circular aperture.
        mask_d : torch.Tensor
            Mask for the central obscuration.
        mask_X : torch.Tensor
            Mask for the horizontal spider.
        mask_Y : torch.Tensor
            Mask for the vertical spider.
        """
        # Coordinates
        x = torch.linspace(-1, 1, resolution, device=self.device)
        X, Y = torch.meshgrid(x, x, indexing='ij')
        d = self.dO * self.D  # Diameter obscuration

        R = torch.sqrt(X**2 + Y**2)

        # Masking image
        mask_D = R > self.D
        mask_d = R < d
        mask_X = torch.abs(X) < self.t / 2
        mask_Y = torch.abs(Y) < self.t / 2

        return mask_D, mask_d, mask_X, mask_Y


def setup_model(config: dict) -> tuple[nn.Module, int]:
    """Returns the model specified in the configuration file, with the last
    layer corresponding to the number of screens.

    Parameters
    ----------
    config : dict
        Dictionary containing the configuration

    Returns
    -------
    model : torch.nn.Module
        The model with the loaded state
    last_model_state : int
        The number of the last model state
    """

    model_name = config['model']['name']
    model_type = config['model']['type']

    print(f'^^^ Initializing model {model_name} of type {model_type}')

    if model_type.startswith('resnet'):
        return get_a_resnet(config)
    elif model_type.startswith('scnnC'):
        return get_scnn(config)
    else:
        raise ValueError(f'Unknown model {model_name}')


def get_a_resnet(config: dict) -> tuple[nn.Module, int]:
    """Returns a pretrained ResNet model, with the last layer corresponding to
    the number of screens.

    Parameters
    ----------
    config : dict
        Dictionary containing the configuration

    Returns
    -------
    model : torch.nn.Module
        The model with the loaded state
    last_model_state : int
        The number of the last model state
    """

    model_name = config['model']['name']
    model_type = config['model']['type']
    pretrained = config['model']['pretrained']
    nscreens = config['speckle']['nscreens']
    data_directory = config['speckle']['datadirectory']
    ensemble = config['preproc'].get('ensemble', 1)

    if model_type == 'resnet18':
        model = torchvision.models.resnet18(
            weights='IMAGENET1K_V1' if pretrained else None)
        finaloutsize = 512
    elif model_type == 'resnet50':
        model = torchvision.models.resnet50(
            weights='IMAGENET1K_V2' if pretrained else None)
        finaloutsize = 2048
    elif model_type == 'resnet152':
        model = torchvision.models.resnet152(
            weights='IMAGENET1K_V2' if pretrained else None)
        finaloutsize = 2048
    else:
        raise ValueError(f'Unknown model {model_type}')

    # If the model uses multiple images as input,
    # add an extra channel as confidence weight
    # to average the final prediction
    if ensemble > 1:
        nscreens = nscreens + 1

    # Give it its name
    model.name = model_name

    # Change the model to process black and white input
    model.conv1 = torch.nn.Conv2d(1,
                                  64,
                                  kernel_size=(7, 7),
                                  stride=(2, 2),
                                  padding=(3, 3),
                                  bias=False)
    # Add a final fully connected piece to predict the output
    model.fc = create_final_block(config, finaloutsize, nscreens)

    return load_model_state(model, data_directory)


def get_scnn(config: dict) -> tuple[nn.Module, int]:
    """Returns a pretrained Spherical-CNN model, with the last layer
    corresponding to the number of screens."""

    model_name = config['model']['name']
    model_type = config['model']['type']
    datadirectory = config['speckle']['datadirectory']

    model_map = {
        'scnnC8': 'C8',
        'scnnC16': 'C16',
        'scnnC4': 'C4',
        'scnnC6': 'C6',
        'scnnC10': 'C10',
        'scnnC12': 'C12',
    }
    try:
        scnn_model = SteerableCNN(config, model_map[model_type])
    except KeyError:
        raise ValueError(f'Unknown model {model_type}')

    scnn_model.name = model_name

    return load_model_state(scnn_model, datadirectory)


class EarlyStopper:
    """ Early stopping to stop the training when the validation loss does not decrease anymore.
    """

    def __init__(self, patience: int = 1, min_delta: float = 0):
        """Initializes the EarlyStopper.

        Parameters
        ----------
        patience: int
            Number of epochs of tolerance before stopping.
        min_delta: float
            Percentage of tolerance in considering the loss acceptable.
        """

        self.patience = patience
        self.min_delta = min_delta
        self.counter = 0
        self.min_validation_loss = float('inf')

    def early_stop(self, validation_loss: float) -> bool:
        """ Computes if the early stop condition is met at the current step.

        Parameters
        ----------
        validation_loss: float
            Current value of the validation loss

        Returns
        -------
        bool
            It returns True if the training has met the stop condition.
        """
        if validation_loss < self.min_validation_loss:
            self.min_validation_loss = validation_loss
            self.counter = 0
        elif validation_loss > self.min_validation_loss * (1 + self.min_delta):
            self.counter += 1
            if self.counter >= self.patience:
                return True
        return False



from __future__ import annotations

from typing import Dict

import matplotlib.pyplot as plt
import numpy as np
import scipy.stats as stats
import torch
from torch import device as Device
from torch import nn

from speckcn2.loss import ComposableLoss
from speckcn2.mlmodels import EnsembleModel
from speckcn2.utils import ensure_directory


def score_plot(
    conf: dict,
    inputs: torch.Tensor,
    tags: list,
    loss: torch.Tensor,
    losses: dict,
    i: int,
    counter: int,
    measures: dict,
    Cn2_pred: torch.Tensor,
    Cn2_true: torch.Tensor,
    recovered_tag_pred: torch.Tensor,
    recovered_tag_true: torch.Tensor,
) -> None:
    """Plots side by side:
    - [0:Nensemble] the input images (single or ensemble)
    - [-3] the predicted/exact tags J
    - [-2] the Cn2 profile
    - [-1] the different information of the loss
    normalize value in model units.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    inputs : torch.Tensor
        The input speckle patterns
    tags : list
        The exact tags of the data
    loss : torch.Tensor
        The total loss of the model (for this prediction)
    losses : dict
        The individual losses of the model
    i : int
        The batch index of the image
    counter : int
        The global index of the image
    measures : dict
        The different measures of the model
    Cn2_pred : torch.Tensor
        The predicted Cn2 profile
    Cn2_true : torch.Tensor
        The true Cn2 profile
    recovered_tag_pred : torch.Tensor
        The predicted tags
    recovered_tag_true : torch.Tensor
        The true tags
    """
    model_name = conf['model']['name']
    data_dir = conf['speckle']['datadirectory']
    ensemble = conf['preproc'].get('ensemble', 1)
    hs = conf['speckle']['splits']
    nscreens = conf['speckle']['nscreens']
    if len(hs) != nscreens:
        print(
            'WARNING: The number of screens does not match the number of splits'
        )
        return

    dirname = f'{data_dir}/{model_name}_score/single-shot_predictions'
    ensure_directory(dirname)

    fig, axs = plt.subplots(1, 3 + ensemble, figsize=(4 * (2 + ensemble), 3.5))

    # (1) Plot the input images
    for n in range(ensemble):
        img = inputs[ensemble * i + n].detach().cpu().squeeze().abs()
        axs[n].imshow(img, cmap='bone')
    title_string = f'Input {ensemble} images' if ensemble > 1 else 'Input single image'
    axs[1].set_title(title_string)

    # (2) Plot J vs nscreens
    axs[-3].plot(recovered_tag_true.squeeze(0).detach().cpu(),
                 'o',
                 label='True')
    axs[-3].plot(recovered_tag_pred.squeeze(0).detach().cpu(),
                 '.',
                 color='tab:red',
                 label='Predicted')
    axs[-3].set_yscale('log')
    axs[-3].set_ylabel('J')
    axs[-3].set_xlabel('# screen')
    axs[-3].legend()

    # (3) Plot Cn2 vs altitude
    axs[-2].plot(hs, Cn2_true.squeeze(0).detach().cpu(), 'o', label='True')
    axs[-2].plot(hs,
                 Cn2_pred.squeeze(0).detach().cpu(),
                 '.',
                 color='tab:red',
                 label='Predicted')
    axs[-2].set_xscale('log')
    axs[-2].set_yscale('log')
    axs[-2].set_ylabel(r'$Cn^2$')
    axs[-2].set_xlabel('Altitude [m]')

    # (4) Plot the recap information
    axs[-1].axis('off')  # Hide axis
    recap_info = f'LOSS TERMS:\nTotal Loss: {loss.item():.4g}\n'
    # the individual losses
    for key, value in losses.items():
        recap_info += f'{key}: {value.item():.4g}\n'
    recap_info += '-------------------\nPARAMETERS:\n'
    # then the single parameters
    for key, value in measures.items():
        recap_info += f'{key}: {value:.4g}\n'
    axs[-1].text(0.5,
                 0.5,
                 recap_info,
                 horizontalalignment='center',
                 verticalalignment='center',
                 fontsize=10,
                 color='black')

    plt.tight_layout()
    plt.savefig(f'{dirname}/single_speckle_loss{loss.item():.4g}.png')
    plt.close()


def plot_loss(conf: dict, model, data_dir: str) -> None:
    """Plots the loss of the model.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    model : torch.nn.Module
        The model to plot the loss of
    data_dir : str
        The directory where the data is stored
    """

    model_name = conf['model']['name']
    data_dir = conf['speckle']['datadirectory']

    dirname = f'{data_dir}/{model_name}_score'
    ensure_directory(dirname)

    fig, axs = plt.subplots(1, 1, figsize=(5, 5))
    axs.plot(model.epoch, model.loss, label='Training loss')
    axs.plot(model.epoch, model.val_loss, label='Validation loss')
    axs.set_xlabel('Epoch')
    axs.set_ylabel('Loss')
    axs.set_yscale('log')
    axs.legend()
    plt.title(f'Model: {model_name}')
    plt.tight_layout()
    plt.savefig(f'{dirname}/loss_{model_name}.png')
    plt.close()


def plot_time(conf: dict, model, data_dir: str) -> None:
    """Plots the time per epoch of the model.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    model : torch.nn.Module
        The model to plot the loss of
    data_dir : str
        The directory where the data is stored
    """

    model_name = conf['model']['name']
    data_dir = conf['speckle']['datadirectory']

    dirname = f'{data_dir}/{model_name}_score'
    ensure_directory(dirname)

    fig, axs = plt.subplots(1, 1, figsize=(5, 5))
    axs.plot(model.epoch, model.time, label='Time per epoch')
    axs.set_xlabel('Epoch')
    axs.set_ylabel('Time [s]')
    axs.legend()
    plt.title(f'Model: {model_name}')
    plt.tight_layout()
    plt.savefig(f'{dirname}/time_{model_name}.png')
    plt.close()


def plot_histo_losses(conf: dict, test_losses: list[dict],
                      data_dir: str) -> None:
    """Plots the histogram of the losses.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    test_losses : list[dict]
        List of all the losses of the test set
    data_dir : str
        The directory where the data is stored
    """
    model_name = conf['model']['name']
    data_dir = conf['speckle']['datadirectory']

    dirname = f'{data_dir}/{model_name}_score/histo_losses'
    ensure_directory(dirname)

    fig, axs = plt.subplots(1, 1, figsize=(5, 5))
    for key in ['MAE', 'Fried', 'Isoplanatic', 'Scintillation_w']:
        loss = [d[key].detach().cpu() for d in test_losses]
        bins = np.logspace(np.log10(min(loss)), np.log10(max(loss)),
                           num=50).tolist()
        axs.hist(loss, bins=bins, alpha=0.5, label=key, density=True)
    axs.set_xlabel('Loss')
    axs.set_ylabel('Frequency')
    axs.set_yscale('log')
    axs.set_xscale('log')
    axs.legend()
    plt.title(f'Model: {model_name}')
    plt.tight_layout()
    plt.savefig(f'{dirname}/histo_losses_{model_name}.png')
    plt.close()


def plot_param_vs_loss(conf: dict,
                       test_losses: list[dict],
                       data_dir: str,
                       measures: list,
                       no_sign: bool = False,
                       nbins: int = 10,
                       linear_bins: bool = False) -> None:
    """Plots the parameter vs the loss. Optionally, it also plots the detailed
    histo for all the bins for the desired metrics.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    test_losses : list[dict]
        List of all the losses of the test set
    data_dir : str
        The directory where the data is stored
    measures : list
        The measures of the model
    no_sign : bool
        If True, it will plot the abs of the relative error
    nbins : int
        The number of bins in which to partition the data
    linear_bins : bool
        If True, the bins are linearly spaced, otherwise they are log spaced
    """
    model_name = conf['model']['name']
    data_dir = conf['speckle']['datadirectory']

    dirname = f'{data_dir}/{model_name}_score'
    ensure_directory(dirname)

    for param, lname, name, units in zip(
        ['Fried_true', 'Isoplanatic_true', 'Scintillation_w_true'],
        ['Fried', 'Isoplanatic', 'Scintillation_w'],
        ['Fried parameter', 'Isoplanatic angle', 'Rytov index'],
        ['[m]', '[rad]', '[1]'],
    ):

        dirname = f'{data_dir}/{model_name}_score'
        p_data = [d[param].detach().cpu() for d in measures]
        if no_sign:
            l_data = [d[lname].detach().cpu() for d in test_losses]
        else:
            pname = lname.split('_true')[0] + '_pred'
            l_data = [((d[pname] - d[param]) / d[param]).detach().cpu()
                      for d in measures]

        pairs = sorted(zip(p_data, l_data))
        params, loss = zip(*pairs)
        params = np.array(params)
        loss = np.array(loss)

        if linear_bins:
            bins = np.linspace(min(params), max(params), num=nbins)
        else:
            bins = np.logspace(np.log10(min(params)),
                               np.log10(max(params)),
                               num=nbins)
        bin_indices = np.digitize(params, bins)
        bin_means = [
            loss[bin_indices == i].mean() if np.any(bin_indices == i) else 0
            for i in range(1, len(bins))
        ]
        bin_stds = [
            loss[bin_indices == i].std() if np.any(bin_indices == i) else 0
            for i in range(1, len(bins))
        ]
        bin_centers = 0.5 * (bins[:-1] + bins[1:])

        # Plotting the results
        fig, axs = plt.subplots(1, 1, figsize=(5, 5))
        axs.errorbar(bin_centers,
                     bin_means,
                     yerr=bin_stds,
                     marker='o',
                     linestyle='-',
                     alpha=0.75)

        # Plot error reference lines+shade
        axs.axhline(y=1.0, linestyle='--', color='tab:red', label='100% error')
        axs.axhline(y=0.5,
                    linestyle='--',
                    color='tab:orange',
                    label='50% error')
        axs.axhline(y=0.1,
                    linestyle='--',
                    color='tab:green',
                    label='10% error')
        axs.axhline(y=-1.0, linestyle='--', color='tab:red')
        axs.axhline(
            y=-0.5,
            linestyle='--',
            color='tab:orange',
        )
        axs.axhline(
            y=-0.1,
            linestyle='--',
            color='tab:green',
        )
        axs.axhline(
            y=0,
            linestyle='--',
            color='black',
        )
        plt.tight_layout()
        x_min, x_max = axs.get_xlim()
        axs.fill_between([x_min, x_max],
                         -0.1,
                         0.1,
                         color='tab:green',
                         alpha=0.1)
        axs.fill_between([x_min, x_max],
                         0.1,
                         0.5,
                         color='tab:orange',
                         alpha=0.1)
        axs.fill_between([x_min, x_max],
                         -0.5,
                         -0.1,
                         color='tab:orange',
                         alpha=0.1)
        axs.fill_between([x_min, x_max], 0.5, 1.0, color='tab:red', alpha=0.1)
        axs.fill_between([x_min, x_max],
                         -1.0,
                         -0.5,
                         color='tab:red',
                         alpha=0.1)
        axs.set_xlabel(f'{name} {units}')
        axs.set_xscale('log')
        axs.set_yscale('symlog', linthresh=0.1)
        axs.set_ylabel('Relative error')
        yticks = [-1, -0.5, -0.1, 0, 0.1, 0.5, 1]
        plt.yticks(yticks)
        yticklabels = ['-100%', '-50%', '-10%', '0', '10%', '50%', '100%']
        plt.gca().set_yticklabels(yticklabels)
        plt.title(f'Model: {model_name}')
        plt.tight_layout()
        plt.savefig(f'{dirname}/{param}_vs_sum_{model_name}.png')
        plt.close()

        # If specified, plot the histogram per single bin
        if conf['preproc'].get(lname + '_details', False):
            print(f'\nComputing {lname} details')
            dirname = f'{data_dir}/{model_name}_score/{lname}_bin_details'
            ensure_directory(dirname)

            for idx, single_bin in enumerate(bin_centers):
                l_data = loss[bin_indices == idx]
                if len(l_data) == 0:
                    continue
                fig, axs = plt.subplots(1, 1, figsize=(5, 5))
                axs.hist(l_data, bins=50, alpha=0.5, density=True)
                mu = np.mean(l_data)
                sigma = np.std(l_data)
                if no_sign:
                    print(
                        'Warning: you are requesting the analysis of absolute value using'
                        + ' normal gaussian assumption.' +
                        'This is not correct and the error will be overestimated.'
                    )
                print(
                    f'{lname} = {single_bin:.3f} -> mu = {mu:.3f}, sigma = {sigma:.3f}'
                )
                if sigma > 0:
                    x = np.linspace(mu - 3 * sigma, mu + 3 * sigma, 100)
                    x = x[x > min(l_data)]
                    x = x[x < max(l_data)]
                    axs.plot(x,
                             stats.norm.pdf(x, mu, sigma),
                             label=f'Average err: {mu:.3f}, Std: {sigma:.3f}')
                axs.set_xlabel(f'Relative error {lname}')
                axs.set_ylabel('Frequency')
                axs.legend()
                plt.title(f'{lname} value = {single_bin:.3f} {units}')
                plt.tight_layout()
                plt.savefig(f'{dirname}/{lname}_bin{idx}.png')
                plt.close()


def plot_param_histo(conf: dict, test_losses: list[dict], data_dir: str,
                     measures: list) -> None:
    """Plots the histograms of different parameters.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    test_losses : list[dict]
        List of all the losses of the test set
    data_dir : str
        The directory where the data is stored
    measures : list
        The measures of the model
    """
    model_name = conf['model']['name']
    data_dir = conf['speckle']['datadirectory']

    dirname = f'{data_dir}/{model_name}_score'
    ensure_directory(dirname)

    for param_model, param_true, name, units in zip(
        ['Fried_pred', 'Isoplanatic_pred', 'Scintillation_w_pred'],
        ['Fried_true', 'Isoplanatic_true', 'Scintillation_w_true'],
        ['Fried parameter', 'Isoplanatic angle', 'Rytov index'],
        ['[m]', '[rad]', '[1]'],
    ):
        fig, axs = plt.subplots(1, 1, figsize=(5, 5))

        params_model = [d[param_model].detach().cpu() for d in measures]
        params_true = [d[param_true].detach().cpu() for d in measures]

        pairs = sorted(zip(params_true, params_model))
        params_true, params_model = zip(*pairs)
        params_true = np.array(params_true)
        params_model = np.array(params_model)

        bins = np.logspace(np.log10(min(params_true)),
                           np.log10(max(params_true)),
                           num=50).tolist()
        axs.hist(params_true,
                 bins=bins,
                 alpha=0.5,
                 label=param_true,
                 density=True)

        bins = np.logspace(np.log10(min(params_model)),
                           np.log10(max(params_model)),
                           num=50).tolist()
        axs.hist(params_model,
                 bins=bins,
                 alpha=0.5,
                 label=param_model,
                 density=True)

        axs.set_xlabel(f'{name} {units}')
        axs.set_xscale('log')
        axs.set_yscale('log')
        axs.set_ylabel('Frequency')
        axs.legend()
        plt.title(f'Model: {model_name}')
        plt.tight_layout()
        plt.savefig(f'{dirname}/histo_{param_true}_{model_name}.png')
        plt.close()


def plot_J_error_details(conf: dict,
                         tags_true: list,
                         tags_pred: list,
                         nbins: int = 10,
                         linear_bins: bool = False) -> None:
    """Function to plot the histograms per single bin of each single screen tag
    to quantify the relative error as a function of J.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    tags_true : list
        The true tags of the validation set
    tags_pred : list
        The predicted tags of the validation set
    nbins : int
        The number of bins in which to partition the data
    linear_bins : bool
        If True, the bins are linearly spaced, otherwise they are log spaced
    """

    nscreens = conf['speckle']['nscreens']
    data_dir = conf['speckle']['datadirectory']
    model_name = conf['model']['name']

    dirname = f'{data_dir}/{model_name}_score/J_bin_details'
    ensure_directory(dirname)

    if conf['preproc'].get('J_details', False):
        for screen_id in range(nscreens):
            print(f'\nComputing screen-{screen_id} details')

            # Collect the data
            params = []
            loss = []
            for i in range(len(tags_true)):
                params.append(tags_true[i][0,
                                           screen_id].detach().cpu().numpy())
                loss.append((
                    (tags_pred[i][0, screen_id] - tags_true[i][0, screen_id]) /
                    (tags_true[i][0, screen_id])).detach().cpu().numpy())
            params = np.array(params)
            loss = np.array(loss)

            if linear_bins:
                bins = np.linspace(min(params), max(params), num=nbins)
            else:
                bins = np.logspace(np.log10(min(params)),
                                   np.log10(max(params)),
                                   num=nbins)
            bin_indices = np.digitize(params, bins)
            # get the average and std of the error per bin of J[screen_id]
            bin_centers = 0.5 * (bins[:-1] + bins[1:])

            for idx, single_bin in enumerate(bin_centers):
                l_data = loss[bin_indices == idx]
                if len(l_data) == 0:
                    continue
                fig, axs = plt.subplots(1, 1, figsize=(5, 5))
                axs.hist(l_data, bins=50, alpha=0.5, density=True)
                mu = np.mean(l_data)
                sigma = np.std(l_data)
                print(
                    f'J-{screen_id} = {single_bin:.3g} -> mu = {mu:.3f}, sigma = {sigma:.3f}'
                )
                if sigma > 0:
                    x = np.linspace(mu - 3 * sigma, mu + 3 * sigma, 100)
                    x = x[x > min(l_data)]
                    x = x[x < max(l_data)]
                    axs.plot(x,
                             stats.norm.pdf(x, mu, sigma),
                             label=f'Average err: {mu:.3f}, Std: {sigma:.3f}')
                axs.set_xlabel(f'Relative error J (screen-{screen_id})')
                axs.set_ylabel('Frequency')
                axs.legend()
                plt.title(f'J (screen-{screen_id}) value = {single_bin:.3g}')
                plt.tight_layout()
                plt.savefig(f'{dirname}/Jscreen{screen_id}_bin{idx}.png')
                plt.close()


def plot_samples_in_ensemble(conf: dict,
                             test_set: list,
                             device: Device,
                             model: nn.Torch,
                             criterion: ComposableLoss,
                             trimming: float = 0.2,
                             n_max_plots: int = 100) -> None:
    """Plot the prediction over a sample and compare it with the ones
    from its ensemble.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    test_set : list
        The test set
    device : torch.device
        The device to use
    model : nn.Torch
        The trained model
    criterion : ComposableLoss
        The loss function
    trimming : float
        The trimming to use for the mean
    n_max_plots : int
        The maximum number of plots
    """

    data_directory = conf['speckle']['datadirectory']
    model_name = conf['model']['name']
    n_screens = conf['speckle']['nscreens']

    dirname = f'{data_directory}/{model_name}_score/single-shot_predictions'
    ensure_directory(dirname)

    # group the sets that have the same n[1]
    grouped_test_set: Dict = {}
    for n in test_set:
        key = tuple(n[1])
        if key not in grouped_test_set:
            grouped_test_set[key] = []
        grouped_test_set[key].append(n)
    print('\nAnalysis of single shot predictions')
    print(f'Number of samples: {len(test_set)}')
    print(f'Number of speckle groups: {len(grouped_test_set)}')
    # Define a random probability to plot each ensemble
    p_plot = n_max_plots / len(grouped_test_set)

    # In the end, we will plot groups that have uncommon values of loss
    loss_min = 1e10
    loss_max = 0
    ensemble_count = 0

    ensemble = EnsembleModel(conf, device)
    with torch.no_grad():
        model.eval()

        for key, value in grouped_test_set.items():
            _outputs = []
            _all_tags_pred = []

            for count, speckle in enumerate(value, 1):
                output, target, _ = ensemble(model, [speckle])
                _outputs.append(output.detach().cpu().numpy())
                loss, losses = criterion(output, target)

                # Get the Cn2 profile and the recovered tags
                Cn2_pred = criterion.reconstruct_cn2(output)
                Cn2_true = criterion.reconstruct_cn2(target)
                recovered_tag_pred = criterion.get_J(output)
                _all_tags_pred.append(recovered_tag_pred.detach().cpu())
                # and get all the measures
                all_measures = criterion._get_all_measures(
                    output, target, Cn2_pred, Cn2_true)

                if count == 1:
                    fig, ax = plt.subplots(1, 3, figsize=(12, 4))
                    loss_0 = loss

                    # (0) Plot the speckle pattern
                    ax[0].axis('off')  # Hide axis
                    ax[0].imshow(speckle[0][0, :, :], cmap='bone')

                    # (1) Plot J vs nscreens
                    recovered_tag_true = criterion.get_J(target)
                    ax[1].plot(recovered_tag_true.squeeze(0).detach().cpu(),
                               '*',
                               label='True',
                               color='tab:green',
                               markersize=10,
                               markeredgecolor='black',
                               zorder=100)
                    ax[1].plot(recovered_tag_pred.squeeze(0).detach().cpu(),
                               'o',
                               label='This speckle',
                               color='tab:red',
                               markersize=7,
                               markeredgecolor='black',
                               zorder=90)

                    # (2) Plot the parameters of this speckle prediction
                    ax[2].axis('off')  # Hide axis
                    recap_info = f'LOSS TERMS:\nTotal Loss: {loss.item():.4g}\n'
                    # the individual losses
                    for key, value in losses.items():
                        recap_info += f'{key}: {value.item():.4g}\n'
                    recap_info += '-------------------\nPARAMETERS:\n'
                    # then the single parameters
                    for key, value in all_measures.items():
                        recap_info += f'{key}: {value:.4g}\n'
                    ax[2].text(0.5,
                               0.5,
                               recap_info,
                               horizontalalignment='center',
                               verticalalignment='center',
                               fontsize=10,
                               color='black')

            # Now at the end of the loop, we decide if this set needs to plotted or not
            # by checking that the loss is uncommon, or via a random probability
            if loss_0 > loss_max or loss_0 < loss_min or np.random.rand(
            ) < p_plot:
                avg_tags_trim = stats.trim_mean(_all_tags_pred,
                                                trimming).squeeze()
                percentiles_50 = np.percentile(_all_tags_pred, [25, 75],
                                               axis=0).squeeze()
                percentiles_68 = np.percentile(_all_tags_pred, [16, 84],
                                               axis=0).squeeze()
                percentiles_95 = np.percentile(_all_tags_pred, [2.5, 97.5],
                                               axis=0).squeeze()

                x_vals = np.arange(n_screens)
                alp = 0.3
                ax[1].plot(avg_tags_trim,
                           label='Mean',
                           color='tab:red',
                           zorder=50)
                ax[1].fill_between(x_vals,
                                   percentiles_50[0],
                                   percentiles_50[1],
                                   color='gold',
                                   alpha=alp,
                                   label='50% CI',
                                   zorder=5)
                ax[1].fill_between(x_vals,
                                   percentiles_68[0],
                                   percentiles_50[0],
                                   color='cadetblue',
                                   alpha=alp,
                                   label='68% CI',
                                   zorder=4)
                ax[1].fill_between(x_vals,
                                   percentiles_50[1],
                                   percentiles_68[1],
                                   color='cadetblue',
                                   alpha=alp,
                                   zorder=4)
                ax[1].fill_between(x_vals,
                                   percentiles_95[0],
                                   percentiles_68[0],
                                   color='blue',
                                   label='95% CI',
                                   alpha=alp,
                                   zorder=3)
                ax[1].fill_between(x_vals,
                                   percentiles_68[1],
                                   percentiles_95[1],
                                   color='blue',
                                   alpha=alp,
                                   zorder=3)

                ax[1].set_yscale('log')
                ax[1].set_ylabel('J')
                ax[1].set_xlabel('# screen')
                ax[1].legend()
                fig.tight_layout()
                plt.subplots_adjust(top=0.92)
                plt.suptitle(
                    'Prediction from a single speckle, compared to similar')
                plt.savefig(
                    f'{dirname}/single_speckle_loss{loss_0.item():.4g}.png')
                loss_max = max(loss_0, loss_max)
                loss_min = min(loss_0, loss_min)
                ensemble_count += 1

            plt.close()

            if ensemble_count >= n_max_plots:
                break



"""This module defines a Steerable Convolutional Neural Network (SteerableCNN)
using the escnn library for equivariant neural networks. The primary class,
SteerableCNN, allows for the creation of a convolutional neural network that
handles various symmetries, making it useful for tasks requiring rotational
invariance.

The module imports necessary libraries, including torch and escnn, and defines utility
functions such as `create_block`, `compute_new_features`, `create_pool`, and `create_final_block`.
These functions help construct convolutional blocks, calculate feature map sizes,
create anti-aliased pooling layers, and build fully connected layers.

The SteerableCNN class initializes with a configuration dictionary and a symmetry parameter,
setting up parameters like kernel sizes, paddings, strides, and feature fields.
It determines the symmetry group and initializes the input type for the network.

The network is built by iterating through specified kernel sizes, creating convolutional
blocks and pooling layers, adding a group pooling layer for invariance, and creating final
fully connected layers using `create_final_block`.

The forward method processes the input tensor through the network, applying each
equivariant block, performing group pooling, and classifying the output using the fully connected layers.

Overall, this module provides a flexible framework for building steerable
convolutional neural networks with configurable symmetries and architectures.
"""
from __future__ import annotations

import torch
from escnn import gspaces, nn


def create_block(in_type, out_type, kernel_size, padding, stride):
    return nn.SequentialModule(
        nn.R2Conv(in_type,
                  out_type,
                  kernel_size=kernel_size,
                  padding=padding,
                  stride=stride,
                  bias=False), nn.InnerBatchNorm(out_type),
        nn.ReLU(out_type, inplace=True))


def compute_new_features(nfeatures, kernel_size, padding, stride):
    # **Out dim of a convolution is:
    # (in_image_res - kernel_size + 2*pad)/stride + 1
    return int((nfeatures - kernel_size + 2 * padding) / stride) + 1


def create_pool(out_type, sigma, padding, stride):
    return nn.SequentialModule(
        nn.PointwiseAvgPoolAntialiased(out_type,
                                       sigma=sigma,
                                       stride=stride,
                                       padding=padding))


def create_final_block(config: dict, n_initial: int,
                       nscreens: int) -> nn.Sequential:
    """Creates a fully connected neural network block based on a predefined
    configuration.

    This function dynamically creates a sequence of PyTorch layers for a fully connected
    neural network. The configuration for the layers is read from a global `config` dictionary
    which should contain a 'final_block' key with a list of layer configurations. Each layer
    configuration is a dictionary that must include a 'type' key with the name of the layer
    class (e.g., 'Linear', 'Dropout', etc.) and can include additional keys for the layer
    parameters.

    The first 'Linear' layer in the configuration has its number of input features set to `n_in`,
    and any 'Linear' layer with 'out_features' set to 'nscreens' has its number of output features
    set to `nscreens`.

    Args:
    Parameters
    ----------
    config : dict
        The global configuration dictionary containing the layer configurations.
    n_initial : int
        The number of input features for the first 'Linear' layer.
    nscreens : int
        The number of output features for any 'Linear' layer with 'out_features' set to 'nscreens'.

    Returns
    ----------
    torch.nn.Sequential: A sequential container of the configured PyTorch layers.
    """
    layers = []

    for layer_config in config['final_block']:
        layer_type = layer_config.pop('type')

        if layer_type == 'Linear':
            # The first linear layer has a constrained number of input features
            if n_initial != -1:
                layer_config['in_features'] = n_initial
                n_initial = -1
            if layer_config['out_features'] == 'nscreens':
                layer_config['out_features'] = nscreens
            # Pass the number of features as args
            n_in = layer_config.pop('in_features')
            n_out = layer_config.pop('out_features')

            layers.append(torch.nn.Linear(n_in, n_out, **layer_config))
        else:
            layer_class = getattr(torch.nn, layer_type)
            layers.append(layer_class(**layer_config))

    return torch.nn.Sequential(*layers)


class SteerableCNN(torch.nn.Module):

    def __init__(self, config, symmetry):

        super(SteerableCNN, self).__init__()

        self.nscreens = config['speckle']['nscreens']
        self.in_image_res = config['preproc']['resize']
        self.ensemble = config['preproc'].get('ensemble', 1)

        self.KERNEL_SIZES = config['scnn']['KERNEL_SIZES']
        self.PADDINGS = config['scnn']['PADDINGS']
        self.STRIDES = config['scnn']['STRIDES']
        self.FEATURE_FIELDS = config['scnn']['FEATURE_FIELDS']
        self.POOL_INDICES = config['scnn']['POOL_INDICES']
        self.SIGMA = config['scnn']['SIGMA']
        self.POOL_STRIDES = config['scnn']['POOL_STRIDES']
        self.POOL_PADDINGS = config['scnn']['POOL_PADDINGS']

        # Decide the symmetry group
        symmetry_map = {
            'C8': 8,
            'C16': 16,
            'C4': 4,
            'C6': 6,
            'C10': 10,
            'C12': 12,
        }
        try:
            self.r2_act = gspaces.rot2dOnR2(N=symmetry_map[symmetry])
        except KeyError:
            raise ValueError('The symmetry must be one of ' +
                             ', '.join(symmetry_map.keys()))

        # Start computing the size of the feature map
        self.nfeatures = self.in_image_res

        # The input image is a scalar field, corresponding to the trivial representation
        in_type = nn.FieldType(self.r2_act, [self.r2_act.trivial_repr])
        # we store the input type for wrapping the images into a geometric tensor during the forward pass
        self.input_type = in_type
        # This mask is to remove the 'corner' pixels in the input image,
        # that would get outside of the square while rotating
        self.mask = nn.MaskModule(in_type, self.in_image_res, margin=0),

        self.blocks = torch.nn.ModuleList()
        for i in range(len(self.KERNEL_SIZES)):
            # Specify the output type of the convolutional layer
            # we choose feature fields, each transforming under the regular representation of the group
            out_type = nn.FieldType(
                self.r2_act,
                self.FEATURE_FIELDS[i] * [self.r2_act.regular_repr])
            block = create_block(in_type, out_type, self.KERNEL_SIZES[i],
                                 self.PADDINGS[i], self.STRIDES[i])
            self.blocks.append(block)
            in_type = block.out_type
            self.nfeatures = compute_new_features(self.nfeatures,
                                                  self.KERNEL_SIZES[i],
                                                  self.PADDINGS[i],
                                                  self.STRIDES[i])
            # Once every two layers, add a pooling layer
            if i in self.POOL_INDICES:
                j = self.POOL_INDICES.index(i)
                pool = create_pool(out_type, self.SIGMA[j],
                                   self.POOL_PADDINGS[j], self.POOL_STRIDES[j])
                self.blocks.append(pool)
                # For gaussian blur filters, the kernel size is 2 * int(round(3 * sigma)) + 1
                kernel_size = 2 * int(round(3 * self.SIGMA[j])) + 1
                self.nfeatures = compute_new_features(self.nfeatures,
                                                      kernel_size,
                                                      self.POOL_PADDINGS[j],
                                                      self.POOL_STRIDES[j])
        if self.nfeatures < 1:
            raise ValueError(
                'The number of features is too small. Please, check the kernel sizes, paddings and strides'
            )

        # Apply a group pooling layer to impose the invariance
        self.gpool = nn.GroupPooling(out_type)

        # number of output channels
        nc = self.gpool.out_type.size * self.nfeatures * self.nfeatures

        # If the model uses multiple images as input,
        # I add an extra channel as confidence weight
        # to average the final prediction
        if self.ensemble > 1:
            out_size = self.nscreens + 1
        else:
            out_size = self.nscreens

        # Create the final fully connected layers
        self.fully_net = create_final_block(config, nc, out_size)

    def forward(self, input: torch.Tensor):
        # wrap the input tensor in a GeometricTensor
        # (associate it with the input type)
        x = nn.GeometricTensor(input, self.input_type)

        # apply each equivariant block
        for block in self.blocks:
            x = block(x)

        # pool over the group
        x = self.gpool(x)

        # unwrap the output GeometricTensor
        # (take the Pytorch tensor and discard the associated representation)
        x = x.tensor

        # classify with the final fully connected layers
        x = self.fully_net(x.reshape(x.shape[0], -1))

        return x
